# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.dae.flatten import flatten_dae_components
from pyomo.common.modeling import NOTSET
from pyomo.core.base.var import Var
from pyomo.core.base.expression import Expression
from pyomo.core.base.componentuid import ComponentUID
from pyomo.core.expr.numeric_expr import value as pyo_value

from pyomo.contrib.mpc.interfaces.load_data import (
    load_data_from_scalar,
    load_data_from_series,
    load_data_from_interval,
)
from pyomo.contrib.mpc.interfaces.copy_values import copy_values_at_time
from pyomo.contrib.mpc.data.find_nearest_index import find_nearest_index
from pyomo.contrib.mpc.data.get_cuid import get_indexed_cuid
from pyomo.contrib.mpc.data.dynamic_data_base import _is_iterable
from pyomo.contrib.mpc.data.series_data import TimeSeriesData
from pyomo.contrib.mpc.data.interval_data import IntervalData
from pyomo.contrib.mpc.data.scalar_data import ScalarData
from pyomo.contrib.mpc.data.convert import _process_to_dynamic_data
from pyomo.contrib.mpc.modeling.cost_expressions import (
    get_penalty_from_constant_target,
    get_penalty_from_target,
)
from pyomo.contrib.mpc.modeling.constraints import get_piecewise_constant_constraints

iterable_scalars = (str, bytes)


def _to_iterable(item):
    if hasattr(item, "__iter__"):
        if isinstance(item, iterable_scalars):
            yield item
        else:
            for obj in item:
                yield obj
    else:
        yield item


class DynamicModelInterface:
    """A helper class for working with dynamic models, e.g. those where
    many components are indexed by some ordered set referred to as "time."

    This class provides methods for interacting with time-indexed
    components, for instance, loading and extracting data or shifting
    values by some time offset. It also provides methods for constructing
    components useful for dynamic optimization.

    """

    def __init__(self, model, time, context=NOTSET):
        """
        Construct with a model and a set. We will flatten the model
        with respect to this set and generate CUIDs with wildcards.

        """
        scalar_vars, dae_vars = flatten_dae_components(model, time, Var)
        scalar_expr, dae_expr = flatten_dae_components(model, time, Expression)
        self.model = model
        self.time = time
        self._scalar_vars = scalar_vars
        self._dae_vars = dae_vars
        self._scalar_expr = scalar_expr
        self._dae_expr = dae_expr

        if context is NOTSET:
            context = model

        # Use buffer to reduce repeated work during name/cuid generation
        cuid_buffer = {}
        self._scalar_var_cuids = [
            ComponentUID(var, cuid_buffer=cuid_buffer, context=context)
            for var in self._scalar_vars
        ]
        self._dae_var_cuids = [
            ComponentUID(var.referent, cuid_buffer=cuid_buffer, context=context)
            for var in self._dae_vars
        ]
        self._dae_expr_cuids = [
            ComponentUID(expr.referent, cuid_buffer=cuid_buffer, context=context)
            for expr in self._dae_expr
        ]

    def get_scalar_variables(self):
        return self._scalar_vars

    def get_indexed_variables(self):
        return self._dae_vars

    def get_scalar_expressions(self):
        return self._scalar_expr

    def get_indexed_expressions(self):
        return self._dae_expr

    def get_scalar_variable_data(self):
        """
        Get data corresponding to non-time-indexed variables.

        Returns
        -------
        dict
            Maps CUIDs of non-time-indexed variables to the value of these
            variables.

        """
        return {
            cuid: var.value
            for cuid, var in zip(self._scalar_var_cuids, self._scalar_vars)
        }

    def get_data_at_time(self, time=None, include_expr=False):
        """
        Gets data at a single time point or set of time points. Note that
        the returned type changes depending on whether a scalar or iterable
        is supplied.

        """
        if time is None:
            # Default is to use the entire time set, treating a singleton
            # as a scalar.
            time = self.time if len(self.time) > 1 else self.time.at(1)
        if _is_iterable(time):
            # Assume time is iterable
            time_list = list(time)
            data = {
                cuid: [var[t].value for t in time]
                for cuid, var in zip(self._dae_var_cuids, self._dae_vars)
            }
            if include_expr:
                data.update(
                    {
                        cuid: [pyo_value(expr[t]) for t in time]
                        for cuid, expr in zip(self._dae_expr_cuids, self._dae_expr)
                    }
                )
            # Return a TimeSeriesData object
            return TimeSeriesData(data, time_list, time_set=self.time)
        else:
            # time is a scalar
            data = {
                cuid: var[time].value
                for cuid, var in zip(self._dae_var_cuids, self._dae_vars)
            }
            if include_expr:
                data.update(
                    {
                        cuid: pyo_value(expr[time])
                        for cuid, expr in zip(self._dae_expr_cuids, self._dae_expr)
                    }
                )
            # Return ScalarData object
            return ScalarData(data)

    def load_data(
        self,
        data,
        time_points=None,
        tolerance=0.0,
        prefer_left=None,
        exclude_left_endpoint=None,
        exclude_right_endpoint=None,
    ):
        """Method to load data into the model.

        Loads data into indicated variables in the model, possibly
        at specified time points.

        Arguments
        ---------
        data: ~scalar_data.ScalarData, TimeSeriesData, or mapping
            If :class:`ScalarData`, loads values into indicated
            variables at all (or specified) time points. If
            :class:`TimeSeriesData`, loads lists of values into time
            points.  If mapping, checks whether each variable and value
            is indexed or iterable and correspondingly loads data into
            variables.

        time_points: Iterable (optional)
            Subset of time points into which data should be loaded.
            Default of None corresponds to loading into all time points.

        """
        if time_points is None:
            time_points = self.time
        data = _process_to_dynamic_data(data, time_set=self.time)

        def _error_if_used(prefer_left, excl_left, excl_right, dtype):
            if any(a is not None for a in (prefer_left, excl_left, excl_right)):
                raise RuntimeError(
                    "prefer_left, exclude_left_endpoint, and exclude_right_endpoint"
                    " can only be set if data is IntervalData-compatible. Got"
                    " prefer_left=%s, exclude_left_endpoint=%s, and"
                    " exclude_right_endpoint=%s while loading data of type %s"
                    % (prefer_left, excl_left, excl_right, dtype)
                )

        excl_left = exclude_left_endpoint
        excl_right = exclude_right_endpoint
        if isinstance(data, ScalarData):
            # This covers the case of non-time-indexed variables
            # as keys.
            _error_if_used(prefer_left, excl_left, excl_right, type(data))
            load_data_from_scalar(data, self.model, time_points)
        elif isinstance(data, TimeSeriesData):
            _error_if_used(prefer_left, excl_left, excl_right, type(data))
            load_data_from_series(data, self.model, time_points, tolerance=tolerance)
        elif isinstance(data, IntervalData):
            prefer_left = True if prefer_left is None else prefer_left
            excl_left = prefer_left if excl_left is None else excl_left
            excl_right = (not prefer_left) if excl_right is None else excl_right
            load_data_from_interval(
                data,
                self.model,
                time_points,
                tolerance=tolerance,
                prefer_left=prefer_left,
                exclude_left_endpoint=excl_left,
                exclude_right_endpoint=excl_right,
            )

    def copy_values_at_time(self, source_time=None, target_time=None):
        """
        Copy values of all time-indexed variables from source time point
        to target time points.

        Parameters
        ----------
        source_time: Float
            Time point from which to copy values.
        target_time: Float or iterable
            Time point or points to which to copy values.

        """
        if source_time is None:
            source_time = self.time.first()
        if target_time is None:
            target_time = self.time
        copy_values_at_time(self._dae_vars, self._dae_vars, source_time, target_time)

    def shift_values_by_time(self, dt):
        """
        Shift values in time indexed variables by a specified time offset.
        """
        seen = set()
        t0 = self.time.first()
        tf = self.time.last()
        time_map = {}
        time_list = list(self.time)
        for var in self._dae_vars:
            if id(var[tf]) in seen:
                # Assume that if var[tf] has been encountered, this is a
                # reference to a "variable" we have already processed.
                continue
            else:
                seen.add(id(var[tf]))
            new_values = []
            for t in time_list:
                if t not in time_map:
                    # Build up a map from target to source time points,
                    # as I don't want to call find_nearest_index more
                    # frequently than I have to.
                    t_new = t + dt
                    idx = find_nearest_index(time_list, t_new, tolerance=None)
                    # If t_new is not a valid time point, we proceed with the
                    # closest valid time point.
                    # We're relying on the fact that indices of t0 or tf are
                    # returned if t_new is outside the bounds of the time set.
                    t_new = time_list[idx]
                    time_map[t] = t_new
                t_new = time_map[t]
                new_values.append(var[t_new].value)
            for i, t in enumerate(self.time):
                var[t].set_value(new_values[i])

    def get_penalty_from_target(
        self,
        target_data,
        time=None,
        variables=None,
        weight_data=None,
        variable_set=None,
        tolerance=None,
        prefer_left=None,
    ):
        """A method to get a quadratic penalty expression from a provided
        setpoint data structure

        Parameters
        ----------
        target_data: ~scalar_data.ScalarData, TimeSeriesData, or IntervalData
            Holds target values for variables
        time: Set (optional)
            Points at which to apply the tracking cost. Default will use
            the model's time set.
        variables: List of Pyomo VarData (optional)
            Subset of variables supplied in setpoint_data to use in the
            tracking cost. Default is to use all variables supplied.
        weight_data: ~scalar_data.ScalarData (optional)
            Holds the weights to use in the tracking cost for each variable
        variable_set: Set (optional)
            A set indexing the list of provided variables, if one already
            exists.
        tolerance: Float (optional)
            Tolerance for checking inclusion in an interval. Only may be
            provided if IntervalData is provided for target_data. In this
            case the default is 0.0.
        prefer_left: Bool (optional)
            Flag indicating whether the left end point of intervals should
            be preferred over the right end point. Only may be provided if
            IntervalData is provided for target_data. In this case the
            default is False.

        Returns
        -------
        Set, Expression
            Set indexing the list of variables to be penalized, and
            Expression indexed by this set and time. This Expression contains
            the weighted tracking cost for each variable at each point in
            time.

        """
        if time is None:
            time = self.time
        target_data = _process_to_dynamic_data(target_data, time_set=self.time)
        if variables is None:
            # Use variables provided by the setpoint.
            # NOTE: Nondeterministic order in non-C Python < 3.7
            # Should these data structures use OrderedDicts internally
            # to enforce an order here?
            variables = [
                self.model.find_component(key) for key in target_data.get_data().keys()
            ]
        else:
            # Variables were provided. These could be anything. Process them
            # to get time-indexed variables on the model.
            variables = [
                self.model.find_component(get_indexed_cuid(var, (self.time,)))
                for var in variables
            ]
        return get_penalty_from_target(
            variables,
            time,
            target_data,
            weight_data=weight_data,
            variable_set=variable_set,
            tolerance=tolerance,
            prefer_left=prefer_left,
        )

    def get_piecewise_constant_constraints(
        self, variables, sample_points, use_next=True, tolerance=0.0
    ):
        """A method to get an indexed constraint ensuring that inputs
        are piecewise constant.

        Parameters
        ----------
        variables: List of Pyomo Vars
            Variables to enforce piecewise constant
        sample_points: List of floats
            Points marking the boundaries of intervals within which
            variables must be constant
        use_next: Bool (optional)
            Whether to enforce constancy by setting each variable equal
            to itself at the next point in time (as opposed to at the
            previous point in time). Default is True.
        tolerance: Float (optional)
            Absolute tolerance used to determine whether provided sample
            points are in the model's time set.

        Returns
        -------
        Tuple:
            First entry is a Set indexing the list of provided variables
            (with integers). Second entry is a constraint indexed by this
            set and time enforcing the piecewise constant condition via
            equality constraints.

        """
        cuids = [get_indexed_cuid(var, (self.time,)) for var in variables]
        variables = [self.model.find_component(cuid) for cuid in cuids]
        time_list = list(self.time)
        # Make sure that sample points exist (within tolerance) in the time
        # set.
        sample_point_indices = [
            find_nearest_index(time_list, t, tolerance=tolerance) for t in sample_points
        ]
        sample_points = [time_list[i] for i in sample_point_indices]
        return get_piecewise_constant_constraints(
            variables, self.time, sample_points, use_next=use_next
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
"""
This module defines the classes that provide an NLP interface based on
the Ampl Solver Library (ASL) implementation
"""

import os
import numpy as np
import logging

from scipy.sparse import coo_matrix, identity
from pyomo.common.deprecation import deprecated
import pyomo.core.base as pyo
from pyomo.common.collections import ComponentMap
from pyomo.common.modeling import unique_component_name
from pyomo.contrib.pynumero.sparse.block_matrix import BlockMatrix
from pyomo.contrib.pynumero.sparse.block_vector import BlockVector
from pyomo.contrib.pynumero.interfaces.nlp import NLP
from pyomo.contrib.pynumero.interfaces.pyomo_nlp import PyomoNLP
from pyomo.contrib.pynumero.interfaces.utils import (
    make_lower_triangular_full,
    CondensedSparseSummation,
)
from pyomo.contrib.pynumero.interfaces.external_grey_box import ExternalGreyBoxBlock
from pyomo.contrib.pynumero.interfaces.nlp_projections import ProjectedNLP
from pyomo.core.base.suffix import SuffixFinder


# Todo: make some of the numpy arrays not writable from __init__
class PyomoNLPWithGreyBoxBlocks(NLP):
    def __init__(self, pyomo_model):
        super(PyomoNLPWithGreyBoxBlocks, self).__init__()

        # get the list of all grey box blocks and build _ExternalGreyBoxAsNLP objects
        greybox_components = []
        # build a map from the names to the variable data objects
        # this is done over *all* variables in active blocks, even
        # if they are not included in this model
        self._pyomo_model_var_names_to_datas = None
        number_of_objectives = 0
        try:
            # We support Pynumero's ExternalGreyBoxBlock modeling
            # objects that are provided through ExternalGreyBoxBlock objects
            # We reclassify these as Pyomo Block objects before building the
            # PyomoNLP object to expose any variables on the block to
            # the underlying Pyomo machinery
            for greybox in pyomo_model.component_objects(
                ExternalGreyBoxBlock, descend_into=True
            ):
                greybox.parent_block().reclassify_component_type(greybox, pyo.Block)
                greybox_components.append(greybox)

            # store the pyomo model
            self._pyomo_model = pyomo_model

            # count the number of objectives excluding grey box objectives
            for obj in self._pyomo_model.component_data_objects(
                pyo.Objective, active=True, descend_into=True
            ):
                number_of_objectives += 1

            # build a PyomoNLP object (will include the "pyomo"
            # part of the model only)
            if number_of_objectives == 0:
                objname = unique_component_name(pyomo_model, "_obj")
                objective = pyomo_model.add_component(objname, pyo.Objective(expr=0.0))
            try:
                self._pyomo_nlp = PyomoNLP(pyomo_model)
            finally:
                if number_of_objectives == 0:
                    pyomo_model.del_component(objective)

            self._pyomo_model_var_names_to_datas = {
                v.getname(fully_qualified=True): v
                for v in pyomo_model.component_data_objects(
                    ctype=pyo.Var, descend_into=True
                )
            }
            self._pyomo_model_constraint_names_to_datas = {
                c.getname(fully_qualified=True): c
                for c in pyomo_model.component_data_objects(
                    ctype=pyo.Constraint, descend_into=True
                )
            }

        finally:
            # Restore the ctypes of the ExternalGreyBoxBlock components
            for greybox in greybox_components:
                greybox.parent_block().reclassify_component_type(
                    greybox, ExternalGreyBoxBlock
                )

        if self._pyomo_nlp.n_primals() == 0:
            raise ValueError(
                "No variables were found in the Pyomo part of the model."
                " PyomoGreyBoxModel requires at least one variable"
                " to be active in a Pyomo objective or constraint"
            )

        # build the list of NLP wrappers for the greybox objects
        greybox_nlps = []
        fixed_vars = []
        for greybox in greybox_components:
            # iterate through the data objects if component is indexed
            for data in greybox.values():
                if data.active:
                    # check that no variables are fixed
                    fixed_vars.extend(v for v in data.inputs.values() if v.fixed)
                    fixed_vars.extend(v for v in data.outputs.values() if v.fixed)
                    greybox_nlp = _ExternalGreyBoxAsNLP(data)
                    if data.get_external_model().has_objective():
                        number_of_objectives += 1
                    greybox_nlps.append(greybox_nlp)

        if number_of_objectives > 1:
            raise ValueError(
                f'Found {number_of_objectives} active objectives. Expected 1.'
            )

        if fixed_vars:
            logging.getLogger(__name__).error(
                'PyomoNLPWithGreyBoxBlocks found fixed variables for the'
                ' inputs and/or outputs of an ExternalGreyBoxBlock. This'
                ' is not currently supported. The fixed variables were:\n\t'
                + '\n\t'.join(f.getname(fully_qualified=True) for f in fixed_vars)
            )
            raise NotImplementedError(
                'PyomoNLPWithGreyBoxBlocks does not support fixed inputs or outputs'
            )

        # let's build up the union of all the primal variables names
        # RBP: Why use names here? Why not just ComponentSet of all
        # data objects?
        primals_names = set(self._pyomo_nlp.primals_names())
        for gbnlp in greybox_nlps:
            primals_names.update(gbnlp.primals_names())

        # sort the names for consistency run to run
        self._n_primals = len(primals_names)
        self._primals_names = primals_names = sorted(primals_names)
        self._pyomo_model_var_datas = [
            self._pyomo_model_var_names_to_datas[nm] for nm in self._primals_names
        ]

        # get the names of all the constraints
        self._constraint_names = list(self._pyomo_nlp.constraint_names())
        self._constraint_datas = [
            self._pyomo_model_constraint_names_to_datas.get(nm)
            for nm in self._constraint_names
        ]
        for gbnlp in greybox_nlps:
            self._constraint_names.extend(gbnlp.constraint_names())
            self._constraint_datas.extend(
                [(gbnlp._block, nm) for nm in gbnlp.constraint_names()]
            )
        self._n_constraints = len(self._constraint_names)

        self._has_hessian_support = True
        for nlp in greybox_nlps:
            if not nlp.has_hessian_support():
                self._has_hessian_support = False

        # wrap all the nlp objects with projected nlp objects
        self._pyomo_nlp = ProjectedNLP(self._pyomo_nlp, primals_names)
        for i, gbnlp in enumerate(greybox_nlps):
            greybox_nlps[i] = ProjectedNLP(greybox_nlps[i], primals_names)

        # build a list of all the nlps in order
        self._nlps = nlps = [self._pyomo_nlp]
        nlps.extend(greybox_nlps)

        # build the primal and dual inits and lb, ub vectors
        self._init_primals = self._pyomo_nlp.init_primals()
        self._primals_lb = self._pyomo_nlp.primals_lb()
        self._primals_ub = self._pyomo_nlp.primals_ub()
        for gbnlp in greybox_nlps:
            local = gbnlp.init_primals()
            mask = ~np.isnan(local)
            self._init_primals[mask] = local[mask]

            local = gbnlp.primals_lb()
            mask = ~np.isnan(local)
            self._primals_lb[mask] = np.maximum(self._primals_lb[mask], local[mask])

            local = gbnlp.primals_ub()
            mask = ~np.isnan(local)
            self._primals_ub[mask] = np.minimum(self._primals_ub[mask], local[mask])

        # all the nan's should be gone (every primal should be initialized)
        if (
            np.any(np.isnan(self._init_primals))
            or np.any(np.isnan(self._primals_lb))
            or np.any(np.isnan(self._primals_ub))
        ):
            raise ValueError(
                'NaN values found in initialization of primals or'
                ' primals_lb or primals_ub in _PyomoNLPWithGreyBoxBlocks.'
            )

        self._init_duals = BlockVector(len(nlps))
        self._dual_values_blockvector = BlockVector(len(nlps))
        self._constraints_lb = BlockVector(len(nlps))
        self._constraints_ub = BlockVector(len(nlps))
        for i, nlp in enumerate(nlps):
            self._init_duals.set_block(i, nlp.init_duals())
            self._constraints_lb.set_block(i, nlp.constraints_lb())
            self._constraints_ub.set_block(i, nlp.constraints_ub())
            self._dual_values_blockvector.set_block(
                i, np.nan * np.zeros(nlp.n_constraints())
            )
        self._init_duals = self._init_duals.flatten()
        self._constraints_lb = self._constraints_lb.flatten()
        self._constraints_ub = self._constraints_ub.flatten()
        # verify that there are no nans in the init_duals
        if (
            np.any(np.isnan(self._init_duals))
            or np.any(np.isnan(self._constraints_lb))
            or np.any(np.isnan(self._constraints_ub))
        ):
            raise ValueError(
                'NaN values found in initialization of duals or'
                ' constraints_lb or constraints_ub in'
                ' _PyomoNLPWithGreyBoxBlocks.'
            )

        self._primal_values = np.nan * np.ones(self._n_primals)
        # set the values of the primals and duals to make sure initial
        # values get all the way through to the underlying models
        self.set_primals(self._init_primals)
        self.set_duals(self._init_duals)
        assert not np.any(np.isnan(self._primal_values))
        assert not np.any(np.isnan(self._dual_values_blockvector))

        # if any of the problem is scaled (i.e., one or more of primals,
        # constraints, or objective), then we want scaling factors for
        # all of them (defaulted to 1)
        need_scaling = False
        # objective is owned by self._pyomo_nlp, not in any of the greybox models
        self._obj_scaling = self._pyomo_nlp.get_obj_scaling()
        if self._obj_scaling is None:
            self._obj_scaling = 1.0
        else:
            need_scaling = True

        scaling_finder = SuffixFinder(
            'scaling_factor', default=1.0, context=self._pyomo_model
        )
        self._primals_scaling = np.fromiter(
            (scaling_finder.find(v) for v in self._pyomo_model_var_datas),
            count=self.n_primals(),
            dtype=float,
        )
        need_scaling = bool(scaling_finder.all_suffixes)

        self._constraints_scaling = BlockVector(len(nlps))
        for i, nlp in enumerate(nlps):
            local_constraints_scaling = nlp.get_constraints_scaling()
            if local_constraints_scaling is None:
                self._constraints_scaling.set_block(i, np.ones(nlp.n_constraints()))
            else:
                self._constraints_scaling.set_block(i, local_constraints_scaling)
                need_scaling = True
        if need_scaling:
            self._constraints_scaling = self._constraints_scaling.flatten()
        else:
            self._obj_scaling = None
            self._primals_scaling = None
            self._constraints_scaling = None

        # compute the jacobian and the hessian to get nnz
        jac = self.evaluate_jacobian()
        self._nnz_jacobian = len(jac.data)

        self._sparse_hessian_summation = None
        self._nnz_hessian_lag = None
        if self._has_hessian_support:
            hess = self.evaluate_hessian_lag()
            self._nnz_hessian_lag = len(hess.data)

    # overloaded from NLP
    def n_primals(self):
        return self._n_primals

    # overloaded from NLP
    def primals_names(self):
        return self._primals_names

    # overloaded from NLP
    def n_constraints(self):
        return self._n_constraints

    # overloaded from NLP
    def constraint_names(self):
        return self._constraint_names

    # overloaded from NLP
    def nnz_jacobian(self):
        return self._nnz_jacobian

    # overloaded from NLP
    def nnz_hessian_lag(self):
        return self._nnz_hessian_lag

    # overloaded from NLP
    def primals_lb(self):
        return self._primals_lb

    # overloaded from NLP
    def primals_ub(self):
        return self._primals_ub

    # overloaded from NLP
    def constraints_lb(self):
        return self._constraints_lb

    # overloaded from NLP
    def constraints_ub(self):
        return self._constraints_ub

    # overloaded from NLP
    def init_primals(self):
        return self._init_primals

    # overloaded from NLP
    def init_duals(self):
        return self._init_duals

    # overloaded from NLP / Extended NLP
    def create_new_vector(self, vector_type):
        if vector_type == 'primals':
            return np.zeros(self.n_primals(), dtype=np.float64)
        elif vector_type == 'constraints' or vector_type == 'duals':
            return np.zeros(self.n_constraints(), dtype=np.float64)
        else:
            raise RuntimeError('Called create_new_vector with an unknown vector_type')

    # overloaded from NLP
    def set_primals(self, primals):
        np.copyto(self._primal_values, primals)
        for nlp in self._nlps:
            nlp.set_primals(primals)

    # overloaded from AslNLP
    def get_primals(self):
        return np.copy(self._primal_values)

    # overloaded from NLP
    def set_duals(self, duals):
        self._dual_values_blockvector.copyfrom(duals)
        for i, nlp in enumerate(self._nlps):
            nlp.set_duals(self._dual_values_blockvector.get_block(i))

    # overloaded from NLP
    def get_duals(self):
        return self._dual_values_blockvector.flatten()

    # overloaded from NLP
    def set_obj_factor(self, obj_factor):
        for nlp in self._nlps:
            nlp.set_obj_factor(obj_factor)

    # overloaded from NLP
    def get_obj_factor(self):
        # objective is owned by the pyomo model
        return self._pyomo_nlp.get_obj_factor()

    # overloaded from NLP
    def get_obj_scaling(self):
        return self._obj_scaling

    # overloaded from NLP
    def get_primals_scaling(self):
        return self._primals_scaling

    # overloaded from NLP
    def get_constraints_scaling(self):
        return self._constraints_scaling

    # overloaded from NLP
    def evaluate_objective(self):
        # There is a check in the constructor to ensure
        # that there is only one objective. The rest of
        # the nlps will return zero.
        obj = 0
        for nlp in self._nlps:
            obj += nlp.evaluate_objective()
        return obj

    # overloaded from NLP
    def evaluate_grad_objective(self, out=None):
        # There is a check in the constructor to ensure
        # that there is only one objective. The rest of
        # the nlps will return arrays of zero.
        ret = np.zeros(self.n_primals(), dtype=float)
        for nlp in self._nlps:
            ret += nlp.evaluate_grad_objective()

        if out is not None:
            ret.copyto(out)
            return out

        return ret

    # overloaded from NLP
    def evaluate_constraints(self, out=None):
        # todo: implement the "out" version more efficiently
        ret = BlockVector(len(self._nlps))
        for i, nlp in enumerate(self._nlps):
            ret.set_block(i, nlp.evaluate_constraints())

        if out is not None:
            ret.copyto(out)
            return out

        return ret.flatten()

    # overloaded from NLP
    def evaluate_jacobian(self, out=None):
        ret = BlockMatrix(len(self._nlps), 1)
        for i, nlp in enumerate(self._nlps):
            ret.set_block(i, 0, nlp.evaluate_jacobian())
        ret = ret.tocoo()

        if out is not None:
            assert np.array_equal(ret.row, out.row)
            assert np.array_equal(ret.col, out.col)
            np.copyto(out.data, ret.data)
            return out
        return ret

    def evaluate_hessian_lag(self, out=None):
        list_of_hessians = [nlp.evaluate_hessian_lag() for nlp in self._nlps]
        if self._sparse_hessian_summation is None:
            # This is assuming that the nonzero structures of Hessians
            # do not change
            self._sparse_hessian_summation = CondensedSparseSummation(list_of_hessians)
        ret = self._sparse_hessian_summation.sum(list_of_hessians)

        if out is not None:
            assert np.array_equal(ret.row, out.row)
            assert np.array_equal(ret.col, out.col)
            np.copyto(out.data, ret.data)
            return out
        return ret

    def report_solver_status(self, status_code, status_message):
        raise NotImplementedError('This is not yet implemented.')

    def load_state_into_pyomo(self, bound_multipliers=None):
        # load the values of the primals into the pyomo
        primals = self.get_primals()
        for value, vardata in zip(primals, self._pyomo_model_var_datas):
            vardata.set_value(value)

        # get the active suffixes
        m = self._pyomo_model
        model_suffixes = dict(pyo.suffix.active_import_suffix_generator(m))

        # we need to correct the sign of the multipliers based on whether or
        # not we are minimizing or maximizing - this is done in the ASL interface
        # for ipopt, but does not appear to be done in cyipopt.
        obj_sign = 1.0
        # since we will assert the number of objective functions,
        # we only focus on active objective function.
        objs = list(
            m.component_data_objects(
                ctype=pyo.Objective, active=True, descend_into=True
            )
        )
        assert len(objs) == 1
        if objs[0].sense == pyo.maximize:
            obj_sign = -1.0

        if 'dual' in model_suffixes:
            model_suffixes['dual'].clear()
            dual_values = self._dual_values_blockvector.flatten()
            for value, t in zip(dual_values, self._constraint_datas):
                if type(t) is tuple:
                    model_suffixes['dual'].setdefault(t[0], {})[t[1]] = (
                        -obj_sign * value
                    )
                else:
                    # t is a constraint data
                    model_suffixes['dual'][t] = -obj_sign * value

        if 'ipopt_zL_out' in model_suffixes:
            model_suffixes['ipopt_zL_out'].clear()
            if bound_multipliers is not None:
                model_suffixes['ipopt_zL_out'].update(
                    zip(self._pyomo_model_var_datas, obj_sign * bound_multipliers[0])
                )
        if 'ipopt_zU_out' in model_suffixes:
            model_suffixes['ipopt_zU_out'].clear()
            if bound_multipliers is not None:
                model_suffixes['ipopt_zU_out'].update(
                    zip(self._pyomo_model_var_datas, -obj_sign * bound_multipliers[1])
                )


def _default_if_none(value, default):
    if value is None:
        return default
    return value


class _ExternalGreyBoxAsNLP(NLP):
    """
    This class takes an ExternalGreyBoxModel and makes it look
    like an NLP so it can be used with other interfaces. Currently,
    the ExternalGreyBoxModel supports objectives and equality
    constraints only, so some of the methods are not appropriate
    and raise exceptions
    """

    def __init__(self, external_grey_box_block):
        self._block = external_grey_box_block
        self._ex_model = external_grey_box_block.get_external_model()
        self._obj_factor = 1.0
        n_inputs = len(self._block.inputs)
        assert n_inputs == self._ex_model.n_inputs()
        n_eq_constraints = self._ex_model.n_equality_constraints()
        n_outputs = len(self._block.outputs)
        assert n_outputs == self._ex_model.n_outputs()

        # create the list of primals and constraint names
        # primals will be ordered inputs, followed by outputs
        self._primals_names = [
            self._block.inputs[k].getname(fully_qualified=True)
            for k in self._block.inputs
        ]
        self._primals_names.extend(
            self._block.outputs[k].getname(fully_qualified=True)
            for k in self._block.outputs
        )
        n_primals = len(self._primals_names)

        prefix = self._block.getname(fully_qualified=True)
        self._constraint_names = [
            '{}.{}'.format(prefix, nm)
            for nm in self._ex_model.equality_constraint_names()
        ]
        output_var_names = [
            self._block.outputs[k].getname(fully_qualified=False)
            for k in self._block.outputs
        ]
        self._constraint_names.extend(
            [
                '{}.output_constraints[{}]'.format(prefix, nm)
                for nm in self._ex_model.output_names()
            ]
        )

        # create the numpy arrays of bounds on the primals
        self._primals_lb = BlockVector(2)
        self._primals_ub = BlockVector(2)
        self._init_primals = BlockVector(2)
        lb = np.nan * np.zeros(n_inputs)
        ub = np.nan * np.zeros(n_inputs)
        init_primals = np.nan * np.zeros(n_inputs)
        for i, k in enumerate(self._block.inputs):
            lb[i] = _default_if_none(self._block.inputs[k].lb, -np.inf)
            ub[i] = _default_if_none(self._block.inputs[k].ub, np.inf)
            init_primals[i] = _default_if_none(self._block.inputs[k].value, 0.0)
        self._primals_lb.set_block(0, lb)
        self._primals_ub.set_block(0, ub)
        self._init_primals.set_block(0, init_primals)

        lb = np.nan * np.zeros(n_outputs)
        ub = np.nan * np.zeros(n_outputs)
        init_primals = np.nan * np.zeros(n_outputs)
        for i, k in enumerate(self._block.outputs):
            lb[i] = _default_if_none(self._block.outputs[k].lb, -np.inf)
            ub[i] = _default_if_none(self._block.outputs[k].ub, np.inf)
            init_primals[i] = _default_if_none(self._block.outputs[k].value, 0.0)
        self._primals_lb.set_block(1, lb)
        self._primals_ub.set_block(1, ub)
        self._init_primals.set_block(1, init_primals)
        self._primals_lb = self._primals_lb.flatten()
        self._primals_ub = self._primals_ub.flatten()
        self._init_primals = self._init_primals.flatten()

        # create a numpy array to store the values of the primals
        self._primal_values = np.copy(self._init_primals)
        # make sure the values are passed through to other objects
        self.set_primals(self._init_primals)

        # create the numpy arrays for the duals and initial values
        # for now, initialize the duals to zero
        self._init_duals = np.zeros(self.n_constraints(), dtype=np.float64)
        # create the numpy arrays to store the dual variables
        self._dual_values = np.copy(self._init_duals)
        # make sure the values are passed through to other objects
        self.set_duals(self._init_duals)

        # create the numpy arrays for bounds on the constraints
        # for now all of these are equalities
        self._constraints_lb = np.zeros(self.n_constraints(), dtype=np.float64)
        self._constraints_ub = np.zeros(self.n_constraints(), dtype=np.float64)

        # do we have hessian support
        self._has_hessian_support = True
        if self._ex_model.n_equality_constraints() > 0 and not hasattr(
            self._ex_model, 'evaluate_hessian_equality_constraints'
        ):
            self._has_hessian_support = False
        if self._ex_model.n_outputs() > 0 and not hasattr(
            self._ex_model, 'evaluate_hessian_outputs'
        ):
            self._has_hessian_support = False
        if self._ex_model.has_objective() and not hasattr(
            self._ex_model, 'evaluate_hessian_objective'
        ):
            self._has_hessian_support = False

        self._nnz_jacobian = None
        self._nnz_hessian_lag = None
        self._cached_constraint_residuals = None
        self._cached_jacobian = None
        self._cached_hessian = None
        self._cached_objective = None
        self._cached_grad_objective = None

    def n_primals(self):
        return len(self._primals_names)

    def primals_names(self):
        return list(self._primals_names)

    def n_constraints(self):
        return len(self._constraint_names)

    def constraint_names(self):
        return list(self._constraint_names)

    def nnz_jacobian(self):
        if self._nnz_jacobian is None:
            J = self.evaluate_jacobian()
            self._nnz_jacobian = len(J.data)
        return self._nnz_jacobian

    def nnz_hessian_lag(self):
        if self._nnz_hessian_lag is None:
            H = self.evaluate_hessian_lag()
            self._nnz_hessian_lag = len(H.data)
        return self._nnz_hessian_lag

    def primals_lb(self):
        return np.copy(self._primals_lb)

    def primals_ub(self):
        return np.copy(self._primals_ub)

    def constraints_lb(self):
        return np.copy(self._constraints_lb)

    def constraints_ub(self):
        return np.copy(self._constraints_ub)

    def init_primals(self):
        return np.copy(self._init_primals)

    def init_duals(self):
        return np.copy(self._init_duals)

    def create_new_vector(self, vector_type):
        if vector_type == 'primals':
            return np.zeros(self.n_primals(), dtype=np.float64)
        elif vector_type == 'constraints' or vector_type == 'duals':
            return np.zeros(self.n_constraints(), dtype=np.float64)

    def _cache_invalidate_primals(self):
        self._cached_constraint_residuals = None
        self._cached_jacobian = None
        self._cached_hessian = None
        self._cached_objective = None
        self._cached_grad_objective = None

    def set_primals(self, primals):
        self._cache_invalidate_primals()
        assert len(primals) == self.n_primals()
        np.copyto(self._primal_values, primals)
        self._ex_model.set_input_values(primals[: self._ex_model.n_inputs()])

    def get_primals(self):
        return np.copy(self._primal_values)

    def _cache_invalidate_duals(self):
        self._cached_hessian = None

    def set_duals(self, duals):
        self._cache_invalidate_duals()
        assert len(duals) == self.n_constraints()
        np.copyto(self._dual_values, duals)
        if self._ex_model.n_equality_constraints() > 0:
            self._ex_model.set_equality_constraint_multipliers(
                self._dual_values[: self._ex_model.n_equality_constraints()]
            )
        if self._ex_model.n_outputs() > 0:
            self._ex_model.set_output_constraint_multipliers(
                self._dual_values[self._ex_model.n_equality_constraints() :]
            )

    def get_duals(self):
        return np.copy(self._dual_values)

    def set_obj_factor(self, obj_factor):
        self._cached_hessian = None
        self._obj_factor = obj_factor

    def get_obj_factor(self):
        return self._obj_factor

    def get_obj_scaling(self):
        raise NotImplementedError(
            '_ExternalGreyBoxAsNLP does not support objective scaling'
        )

    def get_primals_scaling(self):
        raise NotImplementedError(
            '_ExternalGreyBoxAsNLP does not support scaling of primals '
            'directly. This should be handled at a higher level using '
            'suffixes on the Pyomo variables.'
        )

    def get_constraints_scaling(self):
        # todo: would this be better with block vectors
        scaling = np.ones(self.n_constraints(), dtype=np.float64)
        scaled = False
        if self._ex_model.n_equality_constraints() > 0:
            eq_scaling = self._ex_model.get_equality_constraint_scaling_factors()
            if eq_scaling is not None:
                scaling[: self._ex_model.n_equality_constraints()] = eq_scaling
                scaled = True
        if self._ex_model.n_outputs() > 0:
            output_scaling = self._ex_model.get_output_constraint_scaling_factors()
            if output_scaling is not None:
                scaling[self._ex_model.n_equality_constraints() :] = output_scaling
                scaled = True
        if scaled:
            return scaling
        return None

    def _evaluate_objective_if_necessary_and_cache(self):
        if self._ex_model.has_objective():
            if self._cached_objective is None:
                self._cached_objective = self._ex_model.evaluate_objective()
        else:
            self._cached_objective = 0

    def evaluate_objective(self):
        self._evaluate_objective_if_necessary_and_cache()
        return self._cached_objective

    def _evaluate_grad_objective_if_necessary_and_cache(self):
        if self._ex_model.has_objective():
            if self._cached_grad_objective is None:
                self._cached_grad_objective = self._ex_model.evaluate_grad_objective()
        else:
            self._cached_grad_objective = np.zeros(self.n_primals(), dtype=float)

    def evaluate_grad_objective(self, out=None):
        self._evaluate_grad_objective_if_necessary_and_cache()
        if out is not None:
            assert len(out) == self.n_primals()
            np.copyto(out, self._cached_grad_objective)
            return out
        return np.copy(self._cached_grad_objective)

    def _evaluate_constraints_if_necessary_and_cache(self):
        if self._cached_constraint_residuals is None:
            c = BlockVector(2)
            if self._ex_model.n_equality_constraints() > 0:
                c.set_block(0, self._ex_model.evaluate_equality_constraints())
            else:
                c.set_block(0, np.zeros(0, dtype=np.float64))
            if self._ex_model.n_outputs() > 0:
                output_values = self._primal_values[self._ex_model.n_inputs() :]
                c.set_block(1, self._ex_model.evaluate_outputs() - output_values)
            else:
                c.set_block(1, np.zeros(0, dtype=np.float64))
            self._cached_constraint_residuals = c.flatten()

    def evaluate_constraints(self, out=None):
        self._evaluate_constraints_if_necessary_and_cache()
        if out is not None:
            assert len(out) == self.n_constraints()
            np.copyto(out, self._cached_constraint_residuals)
            return out

        return np.copy(self._cached_constraint_residuals)

    def _evaluate_jacobian_if_necessary_and_cache(self):
        if self._cached_jacobian is None:
            jac = BlockMatrix(2, 2)
            jac.set_row_size(0, self._ex_model.n_equality_constraints())
            jac.set_row_size(1, self._ex_model.n_outputs())
            jac.set_col_size(0, self._ex_model.n_inputs())
            jac.set_col_size(1, self._ex_model.n_outputs())

            if self._ex_model.n_equality_constraints() > 0:
                jac.set_block(
                    0, 0, self._ex_model.evaluate_jacobian_equality_constraints()
                )
            if self._ex_model.n_outputs() > 0:
                jac.set_block(1, 0, self._ex_model.evaluate_jacobian_outputs())
                jac.set_block(1, 1, -1.0 * identity(self._ex_model.n_outputs()))

            self._cached_jacobian = jac.tocoo()

    def evaluate_jacobian(self, out=None):
        self._evaluate_jacobian_if_necessary_and_cache()
        if out is not None:
            jac = self._cached_jacobian
            assert np.array_equal(jac.row, out.row)
            assert np.array_equal(jac.col, out.col)
            np.copyto(out.data, jac.data)
            return out

        return self._cached_jacobian.copy()

    def _evaluate_hessian_if_necessary_and_cache(self):
        if self._cached_hessian is None:
            hess = BlockMatrix(2, 2)
            hess.set_row_size(0, self._ex_model.n_inputs())
            hess.set_row_size(1, self._ex_model.n_outputs())
            hess.set_col_size(0, self._ex_model.n_inputs())
            hess.set_col_size(1, self._ex_model.n_outputs())

            n = self._ex_model.n_inputs()
            expected_shape = (n, n)

            # get the hessian of the objective
            if self._ex_model.has_objective():
                obj_hess = (
                    self._ex_model.evaluate_hessian_objective() * self.get_obj_factor()
                )
                if obj_hess.shape != expected_shape:
                    raise ValueError(
                        'ExternalGreyBoxModel objective hessian shape'
                        'should be (n_inputs, n_inputs)'
                    )
                # let's check that it is lower triangular
                if np.any(obj_hess.row < obj_hess.col):
                    raise ValueError(
                        'ExternalGreyBoxModel must return lower'
                        'triangular portion of the Hessian only'
                    )
                obj_hess = make_lower_triangular_full(obj_hess)
            else:
                obj_hess = coo_matrix(([], ([], [])), shape=expected_shape)

            # get the hessian w.r.t. the equality constraints
            if self._ex_model.n_equality_constraints() > 0:
                eq_hess = self._ex_model.evaluate_hessian_equality_constraints()
                if eq_hess.shape != expected_shape:
                    raise ValueError(
                        'ExternalGreyBoxModel equality constraint hessian shape'
                        'should be (n_inputs, n_inputs)'
                    )
                # let's check that it is lower triangular
                if np.any(eq_hess.row < eq_hess.col):
                    raise ValueError(
                        'ExternalGreyBoxModel must return lower '
                        'triangular portion of the Hessian only'
                    )

                eq_hess = make_lower_triangular_full(eq_hess)
            else:
                eq_hess = coo_matrix(([], ([], [])), shape=expected_shape)

            if self._ex_model.n_outputs() > 0:
                output_hess = self._ex_model.evaluate_hessian_outputs()
                if output_hess.shape != expected_shape:
                    raise ValueError(
                        'ExternalGreyBoxModel output hessian shape'
                        'should be (n_inputs, n_inputs)'
                    )
                # let's check that it is lower triangular
                if np.any(output_hess.row < output_hess.col):
                    raise ValueError(
                        'ExternalGreyBoxModel must return lower '
                        'triangular portion of the Hessian only'
                    )

                output_hess = make_lower_triangular_full(output_hess)
            else:
                output_hess = coo_matrix(([], ([], [])), shape=expected_shape)

            row = np.concatenate((obj_hess.row, eq_hess.row, output_hess.row))
            col = np.concatenate((obj_hess.col, eq_hess.col, output_hess.col))
            data = np.concatenate((obj_hess.data, eq_hess.data, output_hess.data))
            input_hess = coo_matrix((data, (row, col)), shape=expected_shape)

            hess.set_block(0, 0, input_hess)
            self._cached_hessian = hess.tocoo()

    def has_hessian_support(self):
        return self._has_hessian_support

    def evaluate_hessian_lag(self, out=None):
        if not self._has_hessian_support:
            raise NotImplementedError(
                'Hessians not supported for all of the external grey box'
                ' models. Therefore, Hessians are not supported overall.'
            )

        self._evaluate_hessian_if_necessary_and_cache()
        if out is not None:
            hess = self._cached_hessian
            assert np.array_equal(hess.row, out.row)
            assert np.array_equal(hess.col, out.col)
            np.copyto(out.data, hess.data)
            return out

        return self._cached_hessian.copy()

    def report_solver_status(self, status_code, status_message):
        raise NotImplementedError('report_solver_status not implemented')



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
"""Functionality for identifying variables that participate in expressions"""

from contextlib import nullcontext

from pyomo.core.expr.visitor import identify_variables
from pyomo.core.expr.numvalue import value as pyo_value
from pyomo.repn import generate_standard_repn
from pyomo.util.subsystems import TemporarySubsystemManager
from pyomo.contrib.incidence_analysis.config import (
    IncidenceMethod,
    get_config_from_kwds,
)


#
# Handlers for different methods of generating the incidence graph
#
def _get_incident_via_identify_variables(expr, include_fixed):
    # Note that identify_variables will not identify the same variable
    # more than once.
    return list(identify_variables(expr, include_fixed=include_fixed))


def _get_incident_via_standard_repn(
    expr, include_fixed, linear_only, compute_values=False
):
    if include_fixed:
        to_unfix = [
            var for var in identify_variables(expr, include_fixed=True) if var.fixed
        ]
        context = TemporarySubsystemManager(to_unfix=to_unfix)
    else:
        context = nullcontext()

    with context:
        repn = generate_standard_repn(
            expr, compute_values=compute_values, quadratic=False
        )

    linear_vars = []
    # Check coefficients to make sure we don't include linear variables with
    # fixed coefficients of zero.
    # Note that linear variables with constant coefficients of zero are already
    # filtered in generate_standard_repn
    for var, coef in zip(repn.linear_vars, repn.linear_coefs):
        try:
            value = pyo_value(coef)
        except ValueError as err:
            # Catch error evaluating expression with uninitialized variables
            # TODO: Suppress logged error?
            if "No value for uninitialized VarData" not in str(err):
                raise err
            value = None
        if value != 0:
            linear_vars.append(var)
    if linear_only:
        nl_var_id_set = set(id(var) for var in repn.nonlinear_vars)
        return [var for var in linear_vars if id(var) not in nl_var_id_set]
    else:
        # Combine linear and nonlinear variables and filter out duplicates. Note
        # that quadratic=False, so we don't need to include repn.quadratic_vars.
        variables = linear_vars + list(repn.nonlinear_vars)
        unique_variables = []
        id_set = set()
        for var in variables:
            v_id = id(var)
            if v_id not in id_set:
                id_set.add(v_id)
                unique_variables.append(var)
        return unique_variables


def _get_incident_via_ampl_repn(expr, linear_only, visitor):
    def _nonlinear_var_id_collector(idlist):
        for _id in idlist:
            if _id in visitor.subexpression_cache:
                info = visitor.subexpression_cache[_id][1]
                if info.nonlinear:
                    yield from _nonlinear_var_id_collector(info.nonlinear[1])
                if info.linear:
                    yield from _nonlinear_var_id_collector(info.linear)
            else:
                yield _id

    var_map = visitor.var_map
    repn = visitor.walk_expression((expr, None, 0, 1.0))

    nonlinear_var_id_set = set()
    unique_nonlinear_var_ids = []
    if repn.nonlinear:
        for v_id in _nonlinear_var_id_collector(repn.nonlinear[1]):
            if v_id not in nonlinear_var_id_set:
                nonlinear_var_id_set.add(v_id)
                unique_nonlinear_var_ids.append(v_id)

    nonlinear_vars = [var_map[v_id] for v_id in unique_nonlinear_var_ids]
    linear_only_vars = [
        var_map[v_id]
        for v_id, coef in repn.linear.items()
        if coef != 0.0 and v_id not in nonlinear_var_id_set
    ]
    if linear_only:
        return linear_only_vars
    else:
        variables = linear_only_vars + nonlinear_vars
        return variables


def get_incident_variables(expr, **kwds):
    """Get variables that participate in an expression

    The exact variables returned depends on the method used to determine incidence.
    For example, ``method=IncidenceMethod.identify_variables`` will return all
    variables participating in the expression, while
    ``method=IncidenceMethod.standard_repn`` will return only the variables
    identified by ``generate_standard_repn`` which ignores variables that only
    appear multiplied by a constant factor of zero.

    Keyword arguments must be valid options for ``IncidenceConfig``.

    Parameters
    ----------
    expr: ``NumericExpression``
        Expression to search for variables

    Returns
    -------
    list of VarData
        List containing the variables that participate in the expression

    Example
    -------

    .. doctest::

       >>> import pyomo.environ as pyo
       >>> from pyomo.contrib.incidence_analysis import get_incident_variables
       >>> m = pyo.ConcreteModel()
       >>> m.x = pyo.Var([1, 2, 3])
       >>> expr = m.x[1] + 2*m.x[2] + 3*m.x[3]**2
       >>> print([v.name for v in get_incident_variables(expr)])
       ['x[1]', 'x[2]', 'x[3]']
       >>> print([v.name for v in get_incident_variables(expr, linear_only=True)])
       ['x[1]', 'x[2]']

    """
    config = get_config_from_kwds(**kwds)
    method = config.method
    include_fixed = config.include_fixed
    linear_only = config.linear_only
    amplrepnvisitor = config._ampl_repn_visitor

    # Check compatibility of arguments
    if linear_only and method is IncidenceMethod.identify_variables:
        raise RuntimeError(
            "linear_only=True is not supported when using identify_variables"
        )
    if include_fixed and method is IncidenceMethod.ampl_repn:
        raise RuntimeError("include_fixed=True is not supported when using ampl_repn")
    if method is IncidenceMethod.ampl_repn and amplrepnvisitor is None:
        # Developer error, this should never happen!
        raise RuntimeError("_ampl_repn_visitor must be provided when using ampl_repn")

    # Dispatch to correct method
    if method is IncidenceMethod.identify_variables:
        return _get_incident_via_identify_variables(expr, include_fixed)
    elif method is IncidenceMethod.standard_repn:
        return _get_incident_via_standard_repn(
            expr, include_fixed, linear_only, compute_values=False
        )
    elif method is IncidenceMethod.standard_repn_compute_values:
        return _get_incident_via_standard_repn(
            expr, include_fixed, linear_only, compute_values=True
        )
    elif method is IncidenceMethod.ampl_repn:
        return _get_incident_via_ampl_repn(expr, linear_only, amplrepnvisitor)
    else:
        raise ValueError(
            f"Unrecognized value {method} for the method used to identify incident"
            f" variables. See the IncidenceMethod enum for valid methods."
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os

from pyomo.common import unittest
from pyomo.common.config import ConfigDict
from pyomo.common.enums import SolverAPIVersion
from pyomo.contrib.solver.common import base


class _LegacyWrappedSolverBase(base.LegacySolverWrapper, base.SolverBase):
    pass


class TestSolverBase(unittest.TestCase):
    def test_class_method_list(self):
        expected_list = [
            'CONFIG',
            'api_version',
            'available',
            'is_persistent',
            'solve',
            'version',
        ]
        method_list = [
            method for method in dir(base.SolverBase) if not method.startswith('_')
        ]
        self.assertEqual(sorted(expected_list), sorted(method_list))

    def test_init(self):
        instance = base.SolverBase()
        self.assertFalse(instance.is_persistent())
        self.assertEqual(instance.name, 'solverbase')
        self.assertEqual(instance.api_version().name, 'V2')
        self.assertEqual(instance.CONFIG, instance.config)
        with self.assertRaises(NotImplementedError):
            self.assertEqual(instance.version(), None)
        with self.assertRaises(NotImplementedError):
            self.assertEqual(instance.solve(None), None)
        with self.assertRaises(NotImplementedError):
            self.assertEqual(instance.available(), None)

    def test_context_manager(self):
        with base.SolverBase() as instance:
            self.assertFalse(instance.is_persistent())
            self.assertEqual(instance.name, 'solverbase')
            self.assertEqual(instance.CONFIG, instance.config)

    def test_config_kwds(self):
        instance = base.SolverBase(tee=True)
        self.assertTrue(instance.config.tee)

    def test_custom_solver_name(self):
        instance = base.SolverBase(name='my_unique_name')
        self.assertEqual(instance.name, 'my_unique_name')


class TestPersistentSolverBase(unittest.TestCase):
    def test_class_method_list(self):
        expected_list = [
            'CONFIG',
            '_get_duals',
            '_get_primals',
            '_get_reduced_costs',
            '_load_vars',
            'add_block',
            'add_constraints',
            'api_version',
            'available',
            'is_persistent',
            'remove_block',
            'remove_constraints',
            'set_instance',
            'set_objective',
            'solve',
            'update_parameters',
            'update_variables',
            'version',
        ]
        method_list = [
            method
            for method in dir(base.PersistentSolverBase)
            if not (method.startswith('__') or method.startswith('_abc'))
        ]
        self.assertEqual(sorted(expected_list), sorted(method_list))

    def test_init(self):
        instance = base.PersistentSolverBase()
        self.assertTrue(instance.is_persistent())
        self.assertEqual(instance.api_version(), SolverAPIVersion.V2)
        with self.assertRaises(NotImplementedError):
            self.assertEqual(instance.set_instance(None), None)
        with self.assertRaises(NotImplementedError):
            self.assertEqual(instance.add_constraints(None), None)
        with self.assertRaises(NotImplementedError):
            self.assertEqual(instance.add_block(None), None)
        with self.assertRaises(NotImplementedError):
            self.assertEqual(instance.remove_constraints(None), None)
        with self.assertRaises(NotImplementedError):
            self.assertEqual(instance.remove_block(None), None)
        with self.assertRaises(NotImplementedError):
            self.assertEqual(instance.set_objective(None), None)
        with self.assertRaises(NotImplementedError):
            self.assertEqual(instance.update_variables(None), None)
        with self.assertRaises(NotImplementedError):
            self.assertEqual(instance.update_parameters(), None)
        with self.assertRaises(NotImplementedError):
            instance._get_primals()
        with self.assertRaises(NotImplementedError):
            instance._get_duals()
        with self.assertRaises(NotImplementedError):
            instance._get_reduced_costs()

    def test_context_manager(self):
        with base.PersistentSolverBase() as instance:
            self.assertTrue(instance.is_persistent())


class TestLegacySolverWrapper(unittest.TestCase):
    def test_class_method_list(self):
        expected_list = [
            'api_version',
            'available',
            'config_block',
            'default_variable_value',
            'license_is_valid',
            'set_options',
            'solve',
            'warm_start_capable',
        ]
        method_list = [
            method
            for method in dir(base.LegacySolverWrapper)
            if not method.startswith('_')
        ]
        self.assertEqual(sorted(expected_list), sorted(method_list))

    def test_context_manager(self):
        with _LegacyWrappedSolverBase() as instance:
            self.assertIsInstance(instance, _LegacyWrappedSolverBase)
            with self.assertRaises(NotImplementedError):
                self.assertFalse(instance.available(False))

    def test_map_config(self):
        # Create a fake/empty config structure that can be added to an empty
        # instance of LegacySolverWrapper
        self.config = ConfigDict(implicit=True)
        self.config.declare(
            'solver_options',
            ConfigDict(implicit=True, description="Options to pass to the solver."),
        )
        instance = _LegacyWrappedSolverBase()
        instance.config = self.config
        instance._map_config(
            True, False, False, 20, True, False, None, None, None, False, None, None
        )
        self.assertTrue(instance.config.tee)
        self.assertFalse(instance.config.load_solutions)
        self.assertEqual(instance.config.time_limit, 20)
        self.assertEqual(instance.config.report_timing, True)
        # Keepfiles should not be created because we did not declare keepfiles on
        # the original config
        with self.assertRaises(AttributeError):
            print(instance.config.keepfiles)
        # We haven't implemented solver_io, suffixes, or logfile
        with self.assertRaises(NotImplementedError):
            instance._map_config(
                False,
                False,
                False,
                20,
                False,
                False,
                None,
                None,
                '/path/to/bogus/file',
                False,
                None,
                None,
            )
        with self.assertRaises(NotImplementedError):
            instance._map_config(
                False,
                False,
                False,
                20,
                False,
                False,
                None,
                '/path/to/bogus/file',
                None,
                False,
                None,
                None,
            )
        with self.assertRaises(NotImplementedError):
            instance._map_config(
                False,
                False,
                False,
                20,
                False,
                False,
                '/path/to/bogus/file',
                None,
                None,
                False,
                None,
                None,
            )
        # If they ask for keepfiles, we redirect them to working_dir
        instance._map_config(
            False, False, False, 20, False, False, None, None, None, True, None, None
        )
        self.assertEqual(instance.config.working_dir, os.getcwd())
        with self.assertRaises(AttributeError):
            print(instance.config.keepfiles)

    def test_solver_options_behavior(self):
        # options can work in multiple ways (set from instantiation, set
        # after instantiation, set during solve).
        # Test case 1: Set at instantiation
        solver = _LegacyWrappedSolverBase(options={'max_iter': 6})
        self.assertEqual(solver.options, {'max_iter': 6})
        self.assertEqual(solver.config.solver_options, {'max_iter': 6})

        # Test case 2: Set later
        solver = _LegacyWrappedSolverBase()
        solver.options = {'max_iter': 4, 'foo': 'bar'}
        self.assertEqual(solver.options, {'max_iter': 4, 'foo': 'bar'})
        self.assertEqual(solver.config.solver_options, {'max_iter': 4, 'foo': 'bar'})

        # Test case 3: pass some options to the mapping (aka, 'solve' command)
        solver = _LegacyWrappedSolverBase()
        solver._map_config(options={'max_iter': 4})
        self.assertEqual(solver.options, {'max_iter': 4})
        self.assertEqual(solver.config.solver_options, {'max_iter': 4})

        # Test case 4: Set at instantiation and override during 'solve' call
        solver = _LegacyWrappedSolverBase(options={'max_iter': 6})
        solver._map_config(options={'max_iter': 4})
        self.assertEqual(solver.options, {'max_iter': 4})
        self.assertEqual(solver.config.solver_options, {'max_iter': 4})

        # solver_options are also supported
        # Test case 1: set at instantiation
        solver = _LegacyWrappedSolverBase(solver_options={'max_iter': 6})
        self.assertEqual(solver.options, {'max_iter': 6})
        self.assertEqual(solver.config.solver_options, {'max_iter': 6})

        # Test case 2: pass some solver_options to the mapping (aka, 'solve' command)
        solver = _LegacyWrappedSolverBase()
        solver._map_config(solver_options={'max_iter': 4})
        self.assertEqual(solver.options, {'max_iter': 4})
        self.assertEqual(solver.config.solver_options, {'max_iter': 4})

        # Test case 3: Set at instantiation and override during 'solve' call
        solver = _LegacyWrappedSolverBase(solver_options={'max_iter': 6})
        solver._map_config(solver_options={'max_iter': 4})
        self.assertEqual(solver.options, {'max_iter': 4})
        self.assertEqual(solver.config.solver_options, {'max_iter': 4})

        # users can mix... sort of
        # Test case 1: Initialize with options, solve with solver_options
        solver = _LegacyWrappedSolverBase(options={'max_iter': 6})
        solver._map_config(solver_options={'max_iter': 4})
        self.assertEqual(solver.options, {'max_iter': 4})
        self.assertEqual(solver.config.solver_options, {'max_iter': 4})

        # users CANNOT initialize both values at the same time, because how
        # do we know what to do with it then?
        # Test case 1: Class instance
        with self.assertRaises(ValueError):
            solver = _LegacyWrappedSolverBase(
                options={'max_iter': 6}, solver_options={'max_iter': 4}
            )
        # Test case 2: Passing to `solve`
        solver = _LegacyWrappedSolverBase()
        with self.assertRaises(ValueError):
            solver._map_config(solver_options={'max_iter': 4}, options={'max_iter': 6})

        # Test that assignment to maps to set_value:
        solver = _LegacyWrappedSolverBase()
        config = ConfigDict(implicit=True)
        config.declare(
            'solver_options',
            ConfigDict(implicit=True, description="Options to pass to the solver."),
        )
        solver.config = config
        solver.config.solver_options.max_iter = 6
        self.assertEqual(solver.options, {'max_iter': 6})
        self.assertEqual(solver.config.solver_options, {'max_iter': 6})

    def test_map_results(self):
        # Unclear how to test this
        pass

    def test_solution_handler(self):
        # Unclear how to test this
        pass



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from collections.abc import MutableSet, Set

from pyomo.common.autoslots import AutoSlots

from ._hasher import hasher


def _rehash_keys(encode, val):
    if encode:
        # TBD [JDS 2/2024]: if we
        #
        # return list(val.values())
        #
        # here, then we get a strange failure when deepcopying
        # ComponentSets containing an _ImplicitAny domain.  We could
        # track it down to the implementation of
        # autoslots.fast_deepcopy, but couldn't find an obvious bug.
        # There is no error if we just return the original dict, or if
        # we return a tuple(val.values)
        return val
    else:
        # object id() may have changed after unpickling,
        # so we rebuild the dictionary keys
        return {hasher[obj.__class__](obj): obj for obj in val.values()}


class ComponentSet(AutoSlots.Mixin, MutableSet):
    """
    This class is a replacement for set that allows Pyomo
    modeling components to be used as entries. The
    underlying hash is based on the Python id() of the
    object, which gets around the problem of hashing
    subclasses of NumericValue. This class is meant for
    creating sets of Pyomo components. The use of non-Pyomo
    components as entries should be avoided (as the behavior
    is undefined).

    References to objects are kept around as long as they
    are entries in the container, so there is no need to
    worry about id() clashes.

    We also override __setstate__ so that we can rebuild the
    container based on possibly updated object ids after
    a deepcopy or pickle.

    *** An instance of this class should never be
    deepcopied/pickled unless it is done so along with
    its component entries (e.g., as part of a block). ***
    """

    __slots__ = ("_data",)
    __autoslot_mappers__ = {"_data": _rehash_keys}
    # Expose a "public" interface to the global _hasher dict
    hasher = hasher

    def __init__(self, iterable=None):
        # maps id_hash(obj) -> obj
        self._data = {}
        if iterable is not None:
            self.update(iterable)

    def __str__(self):
        """String representation of the mapping."""
        tmp = [f"{v} (key={k})" for k, v in self._data.items()]
        return f"ComponentSet({tmp})"

    def update(self, iterable):
        """Update a set with the union of itself and others."""
        if isinstance(iterable, ComponentSet):
            self._data.update(iterable._data)
        else:
            self._data.update((hasher[val.__class__](val), val) for val in iterable)

    #
    # Implement MutableSet abstract methods
    #

    def __contains__(self, val):
        return hasher[val.__class__](val) in self._data

    def __iter__(self):
        return iter(self._data.values())

    def __len__(self):
        return self._data.__len__()

    def add(self, val):
        """Add an element."""
        self._data[hasher[val.__class__](val)] = val

    def discard(self, val):
        """Remove an element. Do not raise an exception if absent."""
        _id = hasher[val.__class__](val)
        if _id in self._data:
            del self._data[_id]

    #
    # Overload MutableSet default implementations
    #

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, Set):
            return False
        return len(self) == len(other) and all(
            hasher[val.__class__](val) in self._data for val in other
        )

    def __ne__(self, other):
        return not (self == other)

    #
    # The remaining MutableSet methods have slow default
    # implementations.
    #

    def clear(self):
        """Remove all elements from this set."""
        self._data.clear()

    def remove(self, val):
        """Remove an element. If not a member, raise a KeyError."""
        try:
            del self._data[hasher[val.__class__](val)]
        except KeyError:
            _id = hasher[val.__class__](val)
            raise KeyError(f"{val} (key={_id})") from None



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import networkx_available

if networkx_available:
    from networkx.classes.digraph import DiGraph
    from networkx.algorithms.traversal.breadth_first_search import bfs_edges
    from networkx.algorithms.bipartite.basic import sets as bipartite_sets
    from networkx.algorithms.bipartite.matching import maximum_matching
    from networkx.algorithms.components import connected_components


"""
This module implements the coarse Dulmage-Mendelsohn partition on a Networkx
bipartite graph. It exists in this common directory as it does not depend on
the rest of the package, or Pyomo except through the import of NetworkX.
"""


def _get_projected_digraph(bg, matching, top_nodes):
    digraph = DiGraph()
    digraph.add_nodes_from(top_nodes)
    for n in top_nodes:
        # Add in-edges
        if n in matching:
            for t in bg[matching[n]]:
                if t != n:
                    digraph.add_edge(t, n)
        # Add out-edges
        for b in bg[n]:
            if b in matching and matching[b] != n:
                digraph.add_edge(n, matching[b])
    return digraph


def _get_reachable_from(digraph, sources):
    _filter = set()
    reachable = []
    for node in sources:
        for i, j in bfs_edges(digraph, node):
            if j not in _filter:
                _filter.add(j)
                reachable.append(j)
    return reachable, _filter


def dulmage_mendelsohn(bg, top_nodes=None, matching=None):
    """
    The Dulmage-Mendelsohn decomposition for bipartite graphs.
    This is the coarse decomposition.
    """
    # TODO: Should top_nodes be required? We can try to infer, but
    # the result is in terms of this partition...
    top, bot = bipartite_sets(bg, top_nodes)
    bot_nodes = [n for n in bg if n not in top]
    if top_nodes is None:
        top_nodes = [n for n in bg if n in top]

    if matching is None:
        # This maps top->bot AND bot->top
        matching = maximum_matching(bg, top_nodes=top_nodes)

    t_unmatched = [t for t in top_nodes if t not in matching]
    b_unmatched = [b for b in bot_nodes if b not in matching]

    # A traversal along these graphs corresponds to an alternating path
    t_digraph = _get_projected_digraph(bg, matching, top_nodes)
    b_digraph = _get_projected_digraph(bg, matching, bot_nodes)

    # Nodes reachable by an alternating path from unmatched nodes
    t_reachable, t_filter = _get_reachable_from(t_digraph, t_unmatched)
    b_reachable, b_filter = _get_reachable_from(b_digraph, b_unmatched)

    # Nodes matched with those reachable from unmatched nodes
    t_matched_with_reachable = [matching[b] for b in b_reachable]
    b_matched_with_reachable = [matching[t] for t in t_reachable]

    _filter = t_filter.union(b_filter)
    _filter.update(t_unmatched)
    _filter.update(t_matched_with_reachable)
    _filter.update(b_unmatched)
    _filter.update(b_matched_with_reachable)
    t_other = [t for t in top_nodes if t not in _filter]
    b_other = [matching[t] for t in t_other]

    return (
        (t_unmatched, t_reachable, t_matched_with_reachable, t_other),
        (b_unmatched, b_reachable, b_matched_with_reachable, b_other),
    )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.opt
import pyomo.environ as pyo
import pyomo.common.unittest as unittest
import pyomo.contrib.alternative_solutions.aos_utils as au
from pyomo.contrib.alternative_solutions import Solution

mip_solver = "gurobi"
mip_available = pyomo.opt.check_available_solvers(mip_solver)


class TestSolutionUnit(unittest.TestCase):

    def get_model(self):
        """
        Simple model with all variable types and fixed variables to test the
        Solution code.
        """
        m = pyo.ConcreteModel()
        m.x = pyo.Var(domain=pyo.NonNegativeReals)
        m.y = pyo.Var(domain=pyo.Binary)
        m.z = pyo.Var(domain=pyo.NonNegativeIntegers)
        m.f = pyo.Var(domain=pyo.Reals)

        m.f.fix(1)
        m.obj = pyo.Objective(expr=m.x + m.y + m.z + m.f, sense=pyo.maximize)

        m.con_x = pyo.Constraint(expr=m.x <= 1.5)
        m.con_y = pyo.Constraint(expr=m.y <= 1)
        m.con_z = pyo.Constraint(expr=m.z <= 3)
        return m

    @unittest.skipUnless(mip_available, "MIP solver not available")
    def test_solution(self):
        """
        Create a Solution Object, call its functions, and ensure the correct
        data is returned.
        """
        model = self.get_model()
        opt = pyo.SolverFactory(mip_solver)
        opt.solve(model)
        all_vars = au.get_model_variables(model, include_fixed=True)

        solution = Solution(model, all_vars, include_fixed=False)
        sol_str = """{
    "fixed_variables": [
        "f"
    ],
    "objective": "obj",
    "objective_value": 6.5,
    "solution": {
        "x": 1.5,
        "y": 1,
        "z": 3
    }
}"""
        assert str(solution) == sol_str

        solution = Solution(model, all_vars)
        sol_str = """{
    "fixed_variables": [
        "f"
    ],
    "objective": "obj",
    "objective_value": 6.5,
    "solution": {
        "f": 1,
        "x": 1.5,
        "y": 1,
        "z": 3
    }
}"""
        assert solution.to_string(round_discrete=True) == sol_str

        sol_val = solution.get_variable_name_values(
            include_fixed=True, round_discrete=True
        )
        self.assertEqual(set(sol_val.keys()), {"x", "y", "z", "f"})
        self.assertEqual(set(solution.get_fixed_variable_names()), {"f"})


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from io import StringIO
import shlex
from tempfile import mkdtemp
import os, sys, math, logging, shutil, time, subprocess

from pyomo.core.base import Constraint, Var, value, Objective
from pyomo.opt import ProblemFormat, SolverFactory

import pyomo.common
from pyomo.common.dependencies import pathlib
from pyomo.common.collections import Bunch
from pyomo.common.tee import TeeStream

from pyomo.opt.base.solvers import _extract_version

from pyomo.core.kernel.block import IBlock
from pyomo.core.kernel.objective import IObjective
from pyomo.core.kernel.variable import IVariable

import pyomo.core.base.suffix
import pyomo.core.kernel.suffix

from pyomo.opt.results import (
    SolverResults,
    SolverStatus,
    Solution,
    SolutionStatus,
    TerminationCondition,
)

from pyomo.common.dependencies import attempt_import
import struct


def _gams_importer():
    try:
        import gams.core.gdx as gdx

        return gdx
    except ImportError:
        try:
            # fall back to the pre-GAMS-45.0 API
            import gdxcc

            return gdxcc
        except:
            # suppress the error from the old API and reraise the current API import error
            pass
        raise


gdxcc, gdxcc_available = attempt_import('gdxcc', importer=_gams_importer)

logger = logging.getLogger('pyomo.solvers')


class _GAMSSolver:
    """Aggregate of common methods for GAMS interfaces"""

    def __init__(self, **kwds):
        self._version = None
        self._default_variable_value = None
        self._metasolver = False

        self._capabilities = Bunch()
        self._capabilities.linear = True
        self._capabilities.quadratic_objective = True
        self._capabilities.quadratic_constraint = True
        self._capabilities.integer = True
        self._capabilities.sos1 = False
        self._capabilities.sos2 = False

        self.options = Bunch()

    def version(self):
        """Returns a 4-tuple describing the solver executable version."""
        if self._version is None:
            self._version = self._get_version()
        return self._version

    def warm_start_capable(self):
        """True is the solver can accept a warm-start solution."""
        return True

    def default_variable_value(self):
        return self._default_variable_value

    def set_options(self, istr):
        if isinstance(istr, str):
            istr = self._options_string_to_dict(istr)
        for key in istr:
            if not istr[key] is None:
                setattr(self.options, key, istr[key])

    @staticmethod
    def _options_string_to_dict(istr):
        ans = {}
        istr = istr.strip()
        if not istr:
            return ans
        if istr[0] == "'" or istr[0] == '"':
            istr = eval(istr)
        tokens = shlex.split(istr)
        for token in tokens:
            index = token.find('=')
            if index == -1:
                raise ValueError(
                    "Solver options must have the form option=value: '%s'" % istr
                )
            try:
                val = eval(token[(index + 1) :])
            except:
                val = token[(index + 1) :]
            ans[token[:index]] = val
        return ans

    def _simple_model(self, n):
        return """
            option limrow = 0;
            option limcol = 0;
            option solprint = off;
            set I / 1 * %s /;
            variables ans;
            positive variables x(I);
            equations obj;
            obj.. ans =g= sum(I, x(I));
            model test / all /;
            solve test using lp minimizing ans;
            """ % (n,)

    #
    # Support "with" statements.
    #
    def __enter__(self):
        return self

    def __exit__(self, t, v, traceback):
        pass


@SolverFactory.register('gams', doc='The GAMS modeling language')
class GAMSSolver(_GAMSSolver):
    """
    A generic interface to GAMS solvers.

    Pass solver_io keyword arg to SolverFactory to choose solver mode:
        solver_io='direct' or 'python' to use GAMS Python API
            Requires installation, visit Python API page on gams.com for help.
        solver_io='shell' or 'gms' to use command line to call gams
            Requires the gams executable be on your system PATH.
    """

    def __new__(cls, *args, **kwds):
        mode = kwds.pop('solver_io', 'shell')
        if mode is None:
            mode = 'shell'

        if mode == 'direct' or mode == 'python':
            return SolverFactory('_gams_direct', **kwds)
        if mode == 'shell' or mode == 'gms':
            return SolverFactory('_gams_shell', **kwds)
        else:
            logger.error('Unknown IO type: %s' % mode)
            return


@SolverFactory.register(
    '_gams_direct', doc='Direct python interface to the GAMS modeling language'
)
class GAMSDirect(_GAMSSolver):
    """
    A generic python interface to GAMS solvers.

    Visit Python API page on gams.com for installation help.
    """

    def available(self, exception_flag=True):
        """True if the solver is available."""
        try:
            from gams import GamsWorkspace, DebugLevel
        except ImportError as e:
            if not exception_flag:
                return False
            raise ImportError(
                "Import of gams failed - GAMS direct "
                "solver functionality is not available.\n"
                "GAMS message: %s" % (e,)
            )
        avail = self._run_simple_model(1)
        if not avail and exception_flag:
            raise NameError(
                "'gams' command failed to solve a simple model - "
                "GAMS shell solver functionality is not available."
            )
        return avail

    def license_is_valid(self):
        # New versions of the community license can run LPs up to 5k
        return self._run_simple_model(5001)

    def _get_version(self):
        """Returns a tuple describing the solver executable version."""
        if not self.available(exception_flag=False):
            return _extract_version('')
        from gams import GamsWorkspace

        workspace = GamsWorkspace()
        version = tuple(int(i) for i in workspace._version.split('.')[:4])
        while len(version) < 4:
            version += (0,)
        return version

    def _run_simple_model(self, n):
        tmpdir = mkdtemp()
        try:
            from gams import GamsWorkspace, DebugLevel

            workspace = GamsWorkspace(debug=DebugLevel.Off, working_directory=tmpdir)
            t1 = workspace.add_job_from_string(self._simple_model(n))
            t1.run()
            return True
        except:
            return False
        finally:
            shutil.rmtree(tmpdir)

    def solve(self, *args, **kwds):
        """
        Solve a model via the GAMS Python API.

        Keyword Arguments
        -----------------
        tee=False: bool
            Output GAMS log to stdout.
        logfile=None: str
            Filename to output GAMS log to a file.
        load_solutions=True: bool
            Load solution into model. If False, the results
            object will contain the solution data.
        keepfiles=False: bool
            Keep temporary files. Equivalent of DebugLevel.KeepFiles.
            Summary of temp files can be found in _gams_py_gjo0.pf
        tmpdir=None: str
            Specify directory path for storing temporary files.
            A directory will be created if one of this name doesn't exist.
            By default uses the system default temporary path.
        report_timing=False: bool
            Print timing reports for presolve, solver, postsolve, etc.
        io_options: dict
            Options that get passed to the writer.
            See writer in pyomo.repn.plugins.gams_writer for details.
            Updated with any other keywords passed to solve method.
        """

        # Make sure available() doesn't crash
        self.available()

        from gams import GamsWorkspace, DebugLevel

        try:
            from gams import GamsExceptionExecution
        except ImportError:
            from gams.workspace import GamsExceptionExecution

        if len(args) != 1:
            raise ValueError(
                'Exactly one model must be passed to solve method of GAMSSolver.'
            )
        model = args[0]

        # self.options are default for each run, overwritten by kwds
        options = dict()
        options.update(self.options)
        options.update(kwds)

        load_solutions = options.pop("load_solutions", True)
        tee = options.pop("tee", False)
        logfile = options.pop("logfile", None)
        keepfiles = options.pop("keepfiles", False)
        tmpdir = options.pop("tmpdir", None)
        report_timing = options.pop("report_timing", False)
        io_options = options.pop("io_options", {})

        # Pass remaining keywords to writer, which will handle
        # any unrecognized arguments
        io_options.update(options)

        initial_time = time.time()

        # Because GAMS changes the CWD when running the solver, we need
        # to convert user-provided file names to absolute paths
        # (relative to the current directory)
        if logfile is not None:
            logfile = os.path.abspath(logfile)

        ####################################################################
        # Presolve
        ####################################################################

        # Create StringIO stream to pass to gams_writer, on which the
        # model file will be written. The writer also passes this StringIO
        # back, but output_file is defined in advance for clarity.
        output_file = StringIO()
        if isinstance(model, IBlock):
            # Kernel blocks have slightly different write method
            smap_id = model.write(
                filename=output_file,
                format=ProblemFormat.gams,
                _called_by_solver=True,
                **io_options,
            )
            symbolMap = getattr(model, "._symbol_maps")[smap_id]
        else:
            _, smap_id = model.write(
                filename=output_file, format=ProblemFormat.gams, io_options=io_options
            )
            symbolMap = model.solutions.symbol_map[smap_id]

        presolve_completion_time = time.time()
        if report_timing:
            print(
                "      %6.2f seconds required for presolve"
                % (presolve_completion_time - initial_time)
            )

        ####################################################################
        # Apply solver
        ####################################################################

        # IMPORTANT - only delete the whole tmpdir if the solver was the one
        # that made the directory. Otherwise, just delete the files the solver
        # made, if not keepfiles. That way the user can select a directory
        # they already have, like the current directory, without having to
        # worry about the rest of the contents of that directory being deleted.
        newdir = True
        if tmpdir is not None and os.path.exists(tmpdir):
            newdir = False

        workspace = GamsWorkspace(
            debug=DebugLevel.KeepFiles if keepfiles else DebugLevel.Off,
            working_directory=tmpdir,
        )

        t1 = workspace.add_job_from_string(output_file.getvalue())

        try:
            with OutputStream(tee=tee, logfile=logfile) as output_stream:
                t1.run(output=output_stream)
        except GamsExceptionExecution as e:
            try:
                if e.rc == 3:
                    # Execution Error
                    check_expr_evaluation(model, symbolMap, 'direct')
            finally:
                # Always name working directory or delete files,
                # regardless of any errors.
                if keepfiles:
                    print(
                        "\nGAMS WORKING DIRECTORY: %s\n" % workspace.working_directory
                    )
                elif tmpdir is not None:
                    # Garbage collect all references to t1.out_db
                    # So that .gdx file can be deleted
                    t1 = rec = rec_lo = rec_hi = None
                    file_removal_gams_direct(tmpdir, newdir)
                raise
        except:
            # Catch other errors and remove files first
            if keepfiles:
                print("\nGAMS WORKING DIRECTORY: %s\n" % workspace.working_directory)
            elif tmpdir is not None:
                # Garbage collect all references to t1.out_db
                # So that .gdx file can be deleted
                t1 = rec = rec_lo = rec_hi = None
                file_removal_gams_direct(tmpdir, newdir)
            raise

        solve_completion_time = time.time()
        if report_timing:
            print(
                "      %6.2f seconds required for solver"
                % (solve_completion_time - presolve_completion_time)
            )

        ####################################################################
        # Postsolve
        ####################################################################

        # import suffixes must be on the top-level model
        if isinstance(model, IBlock):
            model_suffixes = list(
                comp.storage_key
                for comp in pyomo.core.kernel.suffix.import_suffix_generator(
                    model, active=True, descend_into=False
                )
            )
        else:
            model_suffixes = list(
                name
                for (
                    name,
                    comp,
                ) in pyomo.core.base.suffix.active_import_suffix_generator(model)
            )
        extract_dual = 'dual' in model_suffixes
        extract_rc = 'rc' in model_suffixes

        results = SolverResults()
        results.problem.name = os.path.join(
            workspace.working_directory, t1.name + '.gms'
        )
        results.problem.lower_bound = t1.out_db["OBJEST"].find_record().value
        results.problem.upper_bound = t1.out_db["OBJEST"].find_record().value
        results.problem.number_of_variables = t1.out_db["NUMVAR"].find_record().value
        results.problem.number_of_constraints = t1.out_db["NUMEQU"].find_record().value
        results.problem.number_of_nonzeros = t1.out_db["NUMNZ"].find_record().value
        results.problem.number_of_binary_variables = None
        # Includes binary vars:
        results.problem.number_of_integer_variables = (
            t1.out_db["NUMDVAR"].find_record().value
        )
        results.problem.number_of_continuous_variables = (
            t1.out_db["NUMVAR"].find_record().value
            - t1.out_db["NUMDVAR"].find_record().value
        )
        results.problem.number_of_objectives = 1  # required by GAMS writer
        obj = list(model.component_data_objects(Objective, active=True))
        assert len(obj) == 1, 'Only one objective is allowed.'
        obj = obj[0]
        objctvval = t1.out_db["OBJVAL"].find_record().value
        results.problem.sense = obj.sense
        if obj.is_minimizing():
            results.problem.upper_bound = objctvval
        else:
            results.problem.lower_bound = objctvval

        results.solver.name = "GAMS " + str(self.version())

        # Init termination condition to None to give preference to this first
        # block of code, only set certain TC's below if it's still None
        results.solver.termination_condition = None
        results.solver.message = None

        solvestat = t1.out_db["SOLVESTAT"].find_record().value
        if solvestat == 1:
            results.solver.status = SolverStatus.ok
        elif solvestat == 2:
            results.solver.status = SolverStatus.ok
            results.solver.termination_condition = TerminationCondition.maxIterations
        elif solvestat == 3:
            results.solver.status = SolverStatus.ok
            results.solver.termination_condition = TerminationCondition.maxTimeLimit
        elif solvestat == 5:
            results.solver.status = SolverStatus.ok
            results.solver.termination_condition = TerminationCondition.maxEvaluations
        elif solvestat == 7:
            results.solver.status = SolverStatus.aborted
            results.solver.termination_condition = (
                TerminationCondition.licensingProblems
            )
        elif solvestat == 8:
            results.solver.status = SolverStatus.aborted
            results.solver.termination_condition = TerminationCondition.userInterrupt
        elif solvestat == 10:
            results.solver.status = SolverStatus.error
            results.solver.termination_condition = TerminationCondition.solverFailure
        elif solvestat == 11:
            results.solver.status = SolverStatus.error
            results.solver.termination_condition = (
                TerminationCondition.internalSolverError
            )
        elif solvestat == 4:
            results.solver.status = SolverStatus.warning
            results.solver.message = "Solver quit with a problem (see LST file)"
        elif solvestat in (9, 12, 13):
            results.solver.status = SolverStatus.error
        elif solvestat == 6:
            results.solver.status = SolverStatus.unknown

        results.solver.return_code = 0
        # Not sure if this value is actually user time
        # "the elapsed time it took to execute a solve statement in total"
        results.solver.user_time = t1.out_db["ETSOLVE"].find_record().value
        results.solver.system_time = None
        results.solver.wallclock_time = None
        results.solver.termination_message = None

        soln = Solution()

        modelstat = t1.out_db["MODELSTAT"].find_record().value
        if modelstat == 1:
            results.solver.termination_condition = TerminationCondition.optimal
            soln.status = SolutionStatus.optimal
        elif modelstat == 2:
            results.solver.termination_condition = TerminationCondition.locallyOptimal
            soln.status = SolutionStatus.locallyOptimal
        elif modelstat in [3, 18]:
            results.solver.termination_condition = TerminationCondition.unbounded
            soln.status = SolutionStatus.unbounded
        elif modelstat in [4, 5, 6, 10, 19]:
            results.solver.termination_condition = TerminationCondition.infeasible
            soln.status = SolutionStatus.infeasible
        elif modelstat == 7:
            results.solver.termination_condition = TerminationCondition.feasible
            soln.status = SolutionStatus.feasible
        elif modelstat == 8:
            # 'Integer solution model found'
            results.solver.termination_condition = TerminationCondition.optimal
            soln.status = SolutionStatus.optimal
        elif modelstat == 9:
            results.solver.termination_condition = (
                TerminationCondition.intermediateNonInteger
            )
            soln.status = SolutionStatus.other
        elif modelstat == 11:
            # Should be handled above, if modelstat and solvestat both
            # indicate a licensing problem
            if results.solver.termination_condition is None:
                results.solver.termination_condition = (
                    TerminationCondition.licensingProblems
                )
            soln.status = SolutionStatus.error
        elif modelstat in [12, 13]:
            if results.solver.termination_condition is None:
                results.solver.termination_condition = TerminationCondition.error
            soln.status = SolutionStatus.error
        elif modelstat == 14:
            if results.solver.termination_condition is None:
                results.solver.termination_condition = TerminationCondition.noSolution
            soln.status = SolutionStatus.unknown
        elif modelstat in [15, 16, 17]:
            # Having to do with CNS models,
            # not sure what to make of status descriptions
            results.solver.termination_condition = TerminationCondition.optimal
            soln.status = SolutionStatus.unsure
        else:
            # This is just a backup catch, all cases are handled above
            soln.status = SolutionStatus.error

        soln.gap = abs(results.problem.upper_bound - results.problem.lower_bound)

        for sym, obj in symbolMap.bySymbol.items():
            if isinstance(model, IBlock):
                # Kernel variables have no 'parent_component'
                if obj.ctype is IObjective:
                    soln.objective[sym] = {'Value': objctvval}
                if obj.ctype is not IVariable:
                    continue
            else:
                if obj.parent_component().ctype is Objective:
                    soln.objective[sym] = {'Value': objctvval}
                if obj.parent_component().ctype is not Var:
                    continue
            rec = t1.out_db[sym].find_record()
            # obj.value = rec.level
            soln.variable[sym] = {"Value": rec.level}
            if extract_rc and not math.isnan(rec.marginal):
                # Do not set marginals to nan
                # model.rc[obj] = rec.marginal
                soln.variable[sym]['rc'] = rec.marginal

        if extract_dual:
            for c in model.component_data_objects(Constraint, active=True):
                if c.body.is_fixed() or (not (c.has_lb() or c.has_ub())):
                    # the constraint was not sent to GAMS
                    continue
                sym = symbolMap.getSymbol(c)
                if c.equality:
                    rec = t1.out_db[sym].find_record()
                    if not math.isnan(rec.marginal):
                        # model.dual[c] = rec.marginal
                        soln.constraint[sym] = {'dual': rec.marginal}
                    else:
                        # Solver didn't provide marginals,
                        # nothing else to do here
                        break
                else:
                    # Inequality, assume if 2-sided that only
                    # one side's marginal is nonzero
                    # Negate marginal for _lo equations
                    marg = 0
                    if c.lower is not None:
                        rec_lo = t1.out_db[sym + '_lo'].find_record()
                        marg -= rec_lo.marginal
                    if c.upper is not None:
                        rec_hi = t1.out_db[sym + '_hi'].find_record()
                        marg += rec_hi.marginal
                    if not math.isnan(marg):
                        # model.dual[c] = marg
                        soln.constraint[sym] = {'dual': marg}
                    else:
                        # Solver didn't provide marginals,
                        # nothing else to do here
                        break

        results.solution.insert(soln)

        if keepfiles:
            print("\nGAMS WORKING DIRECTORY: %s\n" % workspace.working_directory)
        elif tmpdir is not None:
            # Garbage collect all references to t1.out_db
            # So that .gdx file can be deleted
            t1 = rec = rec_lo = rec_hi = None
            file_removal_gams_direct(tmpdir, newdir)

        ####################################################################
        # Finish with results
        ####################################################################

        results._smap_id = smap_id
        results._smap = None
        if isinstance(model, IBlock):
            if len(results.solution) == 1:
                results.solution(0).symbol_map = getattr(model, "._symbol_maps")[
                    results._smap_id
                ]
                results.solution(0).default_variable_value = (
                    self._default_variable_value
                )
                if load_solutions:
                    model.load_solution(results.solution(0))
            else:
                assert len(results.solution) == 0
            # see the hack in the write method
            # we don't want this to stick around on the model
            # after the solve
            assert len(getattr(model, "._symbol_maps")) == 1
            delattr(model, "._symbol_maps")
            del results._smap_id
            if load_solutions and (len(results.solution) == 0):
                logger.error("No solution is available")
        else:
            if load_solutions:
                model.solutions.load_from(results)
                results._smap_id = None
                results.solution.clear()
            else:
                results._smap = model.solutions.symbol_map[smap_id]
                model.solutions.delete_symbol_map(smap_id)

        postsolve_completion_time = time.time()
        if report_timing:
            print(
                "      %6.2f seconds required for postsolve"
                % (postsolve_completion_time - solve_completion_time)
            )
            print(
                "      %6.2f seconds required total"
                % (postsolve_completion_time - initial_time)
            )

        return results


@SolverFactory.register(
    '_gams_shell', doc='Shell interface to the GAMS modeling language'
)
class GAMSShell(_GAMSSolver):
    """A generic shell interface to GAMS solvers."""

    def available(self, exception_flag=True):
        """True if the solver is available."""
        exe = pyomo.common.Executable("gams")
        if not exe.available():
            if not exception_flag:
                return False
            raise NameError(
                "No 'gams' command found on system PATH - GAMS shell "
                "solver functionality is not available."
            )
        # New versions of GAMS require a license to run anything.
        # Instead of parsing the output, we will try solving a trivial
        # model.
        avail = self._run_simple_model(1)
        if not avail and exception_flag:
            raise NameError(
                "'gams' command failed to solve a simple model - "
                "GAMS shell solver functionality is not available."
            )
        return avail

    def license_is_valid(self):
        # New versions of the community license can run LPs up to 5k
        return self._run_simple_model(5001)

    def _run_simple_model(self, n):
        solver_exec = self.executable()
        if solver_exec is None:
            return False
        tmpdir = mkdtemp()
        try:
            test = os.path.join(tmpdir, 'test.gms')
            with open(test, 'w') as FILE:
                FILE.write(self._simple_model(n))
            result = subprocess.run(
                [solver_exec, test, "curdir=" + tmpdir, 'lo=0'],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            return not result.returncode
        finally:
            shutil.rmtree(tmpdir)
        return False

    def _default_executable(self):
        executable = pyomo.common.Executable("gams")
        if not executable:
            logger.warning(
                "Could not locate the 'gams' executable, "
                "which is required for solver gams"
            )
            self.enable = False
            return None
        return executable.path()

    def executable(self):
        """Returns the executable used by this solver."""
        return self._default_executable()

    def _get_version(self):
        """Returns a tuple describing the solver executable version."""
        solver_exec = self.executable()

        if solver_exec is None:
            return _extract_version('')
        else:
            # specify logging to stdout for windows compatibility
            cmd = [solver_exec, "audit", "lo=3"]
            results = subprocess.run(
                cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
            )
            return _extract_version(results.stdout)

    def _rewrite_path_win8p3(self, path):
        """
        Return the 8.3 short path on Windows; unchanged elsewhere.

        This change is in response to Pyomo/pyomo#3579 which reported
        that GAMS (direct) fails on Windows if there is a space in
        the path. This utility converts paths to their 8.3 short-path version
        (which never have spaces).
        """
        if not sys.platform.startswith("win"):
            return str(path)

        import ctypes, ctypes.wintypes as wt

        GetShortPathNameW = ctypes.windll.kernel32.GetShortPathNameW
        GetShortPathNameW.argtypes = [wt.LPCWSTR, wt.LPWSTR, wt.DWORD]

        # the file must exist, or Windows will not create a short name
        pathlib.Path(path).parent.mkdir(parents=True, exist_ok=True)
        pathlib.Path(path).touch(exist_ok=True)

        buf = ctypes.create_unicode_buffer(260)
        if GetShortPathNameW(str(path), buf, 260):
            return buf.value
        return str(path)

    def solve(self, *args, **kwds):
        """
        Solve a model via the GAMS executable.

        Keyword Arguments
        -----------------
        tee=False: bool
            Output GAMS log to stdout.
        logfile=None: str
            Filename to output GAMS log to a file.
        load_solutions=True: bool
            Load solution into model. If False, the results
            object will contain the solution data.
        keepfiles=False: bool
            Keep temporary files.
        tmpdir=None: str
            Specify directory path for storing temporary files.
            A directory will be created if one of this name doesn't exist.
            By default uses the system default temporary path.
        report_timing=False: bool
            Print timing reports for presolve, solver, postsolve, etc.
        io_options: dict
            Options that get passed to the writer.
            See writer in pyomo.repn.plugins.gams_writer for details.
            Updated with any other keywords passed to solve method.
            Note: put_results is not available for modification on
            GAMSShell solver.
        """

        # Make sure available() doesn't crash
        self.available()

        if len(args) != 1:
            raise ValueError(
                'Exactly one model must be passed to solve method of GAMSSolver.'
            )
        model = args[0]

        # self.options are default for each run, overwritten by kwds
        options = dict()
        options.update(self.options)
        options.update(kwds)

        load_solutions = options.pop("load_solutions", True)
        tee = options.pop("tee", False)
        logfile = options.pop("logfile", None)
        keepfiles = options.pop("keepfiles", False)
        tmpdir = options.pop("tmpdir", None)
        report_timing = options.pop("report_timing", False)
        io_options = options.pop("io_options", {})

        io_options.update(options)

        # Pass remaining keywords to writer, which will handle
        # any unrecognized arguments
        initial_time = time.time()

        # Because GAMS changes the CWD when running the solver, we need
        # to convert user-provided file names to absolute paths
        # (relative to the current directory)
        if logfile is not None:
            logfile = os.path.abspath(logfile)

        ####################################################################
        # Presolve
        ####################################################################

        # IMPORTANT - only delete the whole tmpdir if the solver was the one
        # that made the directory. Otherwise, just delete the files the solver
        # made, if not keepfiles. That way the user can select a directory
        # they already have, like the current directory, without having to
        # worry about the rest of the contents of that directory being deleted.
        newdir = False
        if tmpdir is None:
            tmpdir = mkdtemp()
            newdir = True
        elif not os.path.exists(tmpdir):
            # makedirs creates all necessary intermediate directories in order
            # to create the path to tmpdir, if they don't already exist.
            # However, if keepfiles is False, we only delete the final folder,
            # leaving the rest of the intermediate ones.
            os.makedirs(tmpdir)
            newdir = True

        output = "model.gms"
        output_filename = os.path.join(tmpdir, output)
        lst = "output.lst"
        lst_filename = os.path.join(tmpdir, lst)

        put_results = "results"
        io_options["put_results"] = put_results
        io_options.setdefault("put_results_format", 'gdx' if gdxcc_available else 'dat')

        if io_options['put_results_format'] == 'gdx':
            results_filename = os.path.join(tmpdir, "GAMS_MODEL_p.gdx")
            statresults_filename = os.path.join(tmpdir, "%s_s.gdx" % (put_results,))
        else:
            results_filename = os.path.join(tmpdir, "%s.dat" % (put_results,))
            statresults_filename = os.path.join(tmpdir, "%sstat.dat" % (put_results,))

        if isinstance(model, IBlock):
            # Kernel blocks have slightly different write method
            smap_id = model.write(
                filename=output_filename,
                format=ProblemFormat.gams,
                _called_by_solver=True,
                **io_options,
            )
            symbolMap = getattr(model, "._symbol_maps")[smap_id]
        else:
            _, smap_id = model.write(
                filename=output_filename,
                format=ProblemFormat.gams,
                io_options=io_options,
            )
            symbolMap = model.solutions.symbol_map[smap_id]

        presolve_completion_time = time.time()
        if report_timing:
            print(
                "      %6.2f seconds required for presolve"
                % (presolve_completion_time - initial_time)
            )

        ####################################################################
        # Apply solver
        ####################################################################

        exe = self.executable()
        command = [exe, output, "o=" + lst, "curdir=" + tmpdir]
        if tee and not logfile:
            # default behaviour of gams is to print to console, for
            # compatibility with windows and *nix we want to explicitly log to
            # stdout (see https://www.gams.com/latest/docs/UG_GamsCall.html)
            command.append("lo=3")
        elif not tee and not logfile:
            command.append("lo=0")
        elif not tee and logfile:
            command.append("lo=2")
        elif tee and logfile:
            command.append("lo=4")
        if logfile:
            command.append(f"lf={self._rewrite_path_win8p3(logfile)}")

        try:
            ostreams = [StringIO()]
            if tee:
                ostreams.append(sys.stdout)
            with TeeStream(*ostreams) as t:
                result = subprocess.run(command, stdout=t.STDOUT, stderr=t.STDERR)
            rc = result.returncode
            txt = ostreams[0].getvalue()

            if keepfiles:
                print("\nGAMS WORKING DIRECTORY: %s\n" % tmpdir)

            if rc == 1 or rc == 127:
                raise IOError("Command 'gams' was not recognized")
            elif rc != 0:
                if rc == 3:
                    # Execution Error
                    # Run check_expr_evaluation, which errors if necessary
                    check_expr_evaluation(model, symbolMap, 'shell')
                # If nothing was raised, or for all other cases, raise this
                logger.error(
                    "GAMS encountered an error during solve. "
                    "Check listing file for details."
                )
                logger.error(txt)
                if os.path.exists(lst_filename):
                    with open(lst_filename, 'r') as FILE:
                        logger.error("GAMS Listing file:\n\n%s" % (FILE.read(),))
                raise RuntimeError(
                    "GAMS encountered an error during solve. "
                    "Check listing file for details."
                )

            if io_options['put_results_format'] == 'gdx':
                model_soln, stat_vars = self._parse_gdx_results(
                    results_filename, statresults_filename
                )
            else:
                model_soln, stat_vars = self._parse_dat_results(
                    results_filename, statresults_filename
                )
        finally:
            if not keepfiles:
                if newdir:
                    shutil.rmtree(tmpdir)
                else:
                    os.remove(output_filename)
                    os.remove(lst_filename)
                    os.remove(results_filename)
                    os.remove(statresults_filename)

        solve_completion_time = time.time()
        if report_timing:
            print(
                "      %6.2f seconds required for solver"
                % (solve_completion_time - presolve_completion_time)
            )

        ####################################################################
        # Postsolve
        ####################################################################

        # import suffixes must be on the top-level model
        if isinstance(model, IBlock):
            model_suffixes = list(
                comp.storage_key
                for comp in pyomo.core.kernel.suffix.import_suffix_generator(
                    model, active=True, descend_into=False
                )
            )
        else:
            model_suffixes = list(
                name
                for (
                    name,
                    comp,
                ) in pyomo.core.base.suffix.active_import_suffix_generator(model)
            )
        extract_dual = 'dual' in model_suffixes
        extract_rc = 'rc' in model_suffixes

        results = SolverResults()
        results.problem.name = output_filename
        results.problem.lower_bound = stat_vars["OBJEST"]
        results.problem.upper_bound = stat_vars["OBJEST"]
        results.problem.number_of_variables = stat_vars["NUMVAR"]
        results.problem.number_of_constraints = stat_vars["NUMEQU"]
        results.problem.number_of_nonzeros = stat_vars["NUMNZ"]
        results.problem.number_of_binary_variables = None
        # Includes binary vars:
        results.problem.number_of_integer_variables = stat_vars["NUMDVAR"]
        results.problem.number_of_continuous_variables = (
            stat_vars["NUMVAR"] - stat_vars["NUMDVAR"]
        )
        results.problem.number_of_objectives = 1  # required by GAMS writer
        obj = list(model.component_data_objects(Objective, active=True))
        assert len(obj) == 1, 'Only one objective is allowed.'
        obj = obj[0]
        objctvval = stat_vars["OBJVAL"]
        results.problem.sense = obj.sense
        if obj.is_minimizing():
            results.problem.upper_bound = objctvval
        else:
            results.problem.lower_bound = objctvval

        results.solver.name = "GAMS " + str(self.version())

        # Init termination condition to None to give preference to this first
        # block of code, only set certain TC's below if it's still None
        results.solver.termination_condition = None
        results.solver.message = None

        solvestat = stat_vars["SOLVESTAT"]
        if solvestat == 1:
            results.solver.status = SolverStatus.ok
        elif solvestat == 2:
            results.solver.status = SolverStatus.ok
            results.solver.termination_condition = TerminationCondition.maxIterations
        elif solvestat == 3:
            results.solver.status = SolverStatus.ok
            results.solver.termination_condition = TerminationCondition.maxTimeLimit
        elif solvestat == 5:
            results.solver.status = SolverStatus.ok
            results.solver.termination_condition = TerminationCondition.maxEvaluations
        elif solvestat == 7:
            results.solver.status = SolverStatus.aborted
            results.solver.termination_condition = (
                TerminationCondition.licensingProblems
            )
        elif solvestat == 8:
            results.solver.status = SolverStatus.aborted
            results.solver.termination_condition = TerminationCondition.userInterrupt
        elif solvestat == 10:
            results.solver.status = SolverStatus.error
            results.solver.termination_condition = TerminationCondition.solverFailure
        elif solvestat == 11:
            results.solver.status = SolverStatus.error
            results.solver.termination_condition = (
                TerminationCondition.internalSolverError
            )
        elif solvestat == 4:
            results.solver.status = SolverStatus.warning
            results.solver.message = "Solver quit with a problem (see LST file)"
        elif solvestat in (9, 12, 13):
            results.solver.status = SolverStatus.error
        elif solvestat == 6:
            results.solver.status = SolverStatus.unknown

        results.solver.return_code = rc  # 0
        # Not sure if this value is actually user time
        # "the elapsed time it took to execute a solve statement in total"
        results.solver.user_time = stat_vars["ETSOLVE"]
        results.solver.system_time = None
        results.solver.wallclock_time = None
        results.solver.termination_message = None

        soln = Solution()

        modelstat = stat_vars["MODELSTAT"]
        if modelstat == 1:
            results.solver.termination_condition = TerminationCondition.optimal
            soln.status = SolutionStatus.optimal
        elif modelstat == 2:
            results.solver.termination_condition = TerminationCondition.locallyOptimal
            soln.status = SolutionStatus.locallyOptimal
        elif modelstat in [3, 18]:
            results.solver.termination_condition = TerminationCondition.unbounded
            soln.status = SolutionStatus.unbounded
        elif modelstat in [4, 5, 6, 10, 19]:
            results.solver.termination_condition = TerminationCondition.infeasible
            soln.status = SolutionStatus.infeasible
        elif modelstat == 7:
            results.solver.termination_condition = TerminationCondition.feasible
            soln.status = SolutionStatus.feasible
        elif modelstat == 8:
            # 'Integer solution model found'
            results.solver.termination_condition = TerminationCondition.optimal
            soln.status = SolutionStatus.optimal
        elif modelstat == 9:
            results.solver.termination_condition = (
                TerminationCondition.intermediateNonInteger
            )
            soln.status = SolutionStatus.other
        elif modelstat == 11:
            # Should be handled above, if modelstat and solvestat both
            # indicate a licensing problem
            if results.solver.termination_condition is None:
                results.solver.termination_condition = (
                    TerminationCondition.licensingProblems
                )
            soln.status = SolutionStatus.error
        elif modelstat in [12, 13]:
            if results.solver.termination_condition is None:
                results.solver.termination_condition = TerminationCondition.error
            soln.status = SolutionStatus.error
        elif modelstat == 14:
            if results.solver.termination_condition is None:
                results.solver.termination_condition = TerminationCondition.noSolution
            soln.status = SolutionStatus.unknown
        elif modelstat in [15, 16, 17]:
            # Having to do with CNS models,
            # not sure what to make of status descriptions
            results.solver.termination_condition = TerminationCondition.optimal
            soln.status = SolutionStatus.unsure
        else:
            # This is just a backup catch, all cases are handled above
            soln.status = SolutionStatus.error

        soln.gap = abs(results.problem.upper_bound - results.problem.lower_bound)

        has_rc_info = True
        for sym, obj in symbolMap.bySymbol.items():
            if isinstance(model, IBlock):
                # Kernel variables have no 'parent_component'
                if obj.ctype is IObjective:
                    soln.objective[sym] = {'Value': objctvval}
                if obj.ctype is not IVariable:
                    continue
            else:
                if obj.parent_component().ctype is Objective:
                    soln.objective[sym] = {'Value': objctvval}
                if obj.parent_component().ctype is not Var:
                    continue
            try:
                rec = model_soln[sym]
            except KeyError:
                # no solution returned
                rec = (float('nan'), float('nan'))
            # obj.value = float(rec[0])
            soln.variable[sym] = {"Value": float(rec[0])}
            if extract_rc and has_rc_info:
                try:
                    # model.rc[obj] = float(rec[1])
                    soln.variable[sym]['rc'] = float(rec[1])
                except ValueError:
                    # Solver didn't provide marginals
                    has_rc_info = False

        if extract_dual:
            for c in model.component_data_objects(Constraint, active=True):
                if (c.body.is_fixed()) or (not (c.has_lb() or c.has_ub())):
                    # the constraint was not sent to GAMS
                    continue
                sym = symbolMap.getSymbol(c)
                if c.equality:
                    try:
                        rec = model_soln[sym]
                    except KeyError:
                        # no solution returned
                        rec = (float('nan'), float('nan'))
                    try:
                        # model.dual[c] = float(rec[1])
                        soln.constraint[sym] = {'dual': float(rec[1])}
                    except ValueError:
                        # Solver didn't provide marginals
                        # nothing else to do here
                        break
                else:
                    # Inequality, assume if 2-sided that only
                    # one side's marginal is nonzero
                    # Negate marginal for _lo equations
                    marg = 0
                    if c.lower is not None:
                        try:
                            rec_lo = model_soln[sym + '_lo']
                        except KeyError:
                            # no solution returned
                            rec_lo = (float('nan'), float('nan'))
                        try:
                            marg -= float(rec_lo[1])
                        except ValueError:
                            # Solver didn't provide marginals
                            marg = float('nan')
                    if c.upper is not None:
                        try:
                            rec_hi = model_soln[sym + '_hi']
                        except KeyError:
                            # no solution returned
                            rec_hi = (float('nan'), float('nan'))
                        try:
                            marg += float(rec_hi[1])
                        except ValueError:
                            # Solver didn't provide marginals
                            marg = float('nan')
                    if not math.isnan(marg):
                        # model.dual[c] = marg
                        soln.constraint[sym] = {'dual': marg}
                    else:
                        # Solver didn't provide marginals
                        # nothing else to do here
                        break

        results.solution.insert(soln)

        ####################################################################
        # Finish with results
        ####################################################################

        results._smap_id = smap_id
        results._smap = None
        if isinstance(model, IBlock):
            if len(results.solution) == 1:
                results.solution(0).symbol_map = getattr(model, "._symbol_maps")[
                    results._smap_id
                ]
                results.solution(0).default_variable_value = (
                    self._default_variable_value
                )
                if load_solutions:
                    model.load_solution(results.solution(0))
            else:
                assert len(results.solution) == 0
            # see the hack in the write method
            # we don't want this to stick around on the model
            # after the solve
            assert len(getattr(model, "._symbol_maps")) == 1
            delattr(model, "._symbol_maps")
            del results._smap_id
            if load_solutions and (len(results.solution) == 0):
                logger.error("No solution is available")
        else:
            if load_solutions:
                model.solutions.load_from(results)
                results._smap_id = None
                results.solution.clear()
            else:
                results._smap = model.solutions.symbol_map[smap_id]
                model.solutions.delete_symbol_map(smap_id)

        postsolve_completion_time = time.time()
        if report_timing:
            print(
                "      %6.2f seconds required for postsolve"
                % (postsolve_completion_time - solve_completion_time)
            )
            print(
                "      %6.2f seconds required total"
                % (postsolve_completion_time - initial_time)
            )

        return results

    def _parse_gdx_results(self, results_filename, statresults_filename):
        model_soln = dict()
        stat_vars = dict.fromkeys(
            [
                'MODELSTAT',
                'SOLVESTAT',
                'OBJEST',
                'OBJVAL',
                'NUMVAR',
                'NUMEQU',
                'NUMDVAR',
                'NUMNZ',
                'ETSOLVE',
            ]
        )

        pgdx = gdxcc.new_gdxHandle_tp()
        ret = gdxcc.gdxCreateD(pgdx, os.path.dirname(self.executable()), 128)
        if not ret[0]:
            raise RuntimeError("GAMS GDX failure (gdxCreate): %s." % ret[1])

        if os.path.exists(statresults_filename):
            ret = gdxcc.gdxOpenRead(pgdx, statresults_filename)
            if not ret[0]:
                raise RuntimeError("GAMS GDX failure (gdxOpenRead): %d." % ret[1])

            specVals = gdxcc.doubleArray(gdxcc.GMS_SVIDX_MAX)
            rc = gdxcc.gdxGetSpecialValues(pgdx, specVals)

            specVals[gdxcc.GMS_SVIDX_EPS] = sys.float_info.min
            specVals[gdxcc.GMS_SVIDX_UNDEF] = float("nan")
            specVals[gdxcc.GMS_SVIDX_PINF] = float("inf")
            specVals[gdxcc.GMS_SVIDX_MINF] = float("-inf")
            specVals[gdxcc.GMS_SVIDX_NA] = struct.unpack(
                ">d", bytes.fromhex("fffffffffffffffe")
            )[0]
            gdxcc.gdxSetSpecialValues(pgdx, specVals)

            i = 0
            while True:
                i += 1
                ret = gdxcc.gdxDataReadRawStart(pgdx, i)
                if not ret[0]:
                    break

                ret = gdxcc.gdxSymbolInfo(pgdx, i)
                if not ret[0]:
                    break
                if len(ret) < 2:
                    raise RuntimeError("GAMS GDX failure (gdxSymbolInfo).")
                stat = ret[1]
                if not stat in stat_vars:
                    continue

                ret = gdxcc.gdxDataReadRaw(pgdx)
                if not ret[0] or len(ret[2]) == 0:
                    raise RuntimeError("GAMS GDX failure (gdxDataReadRaw).")

                if stat in ('OBJEST', 'OBJVAL', 'ETSOLVE'):
                    stat_vars[stat] = ret[2][0]
                else:
                    stat_vars[stat] = int(ret[2][0])

            gdxcc.gdxDataReadDone(pgdx)
            gdxcc.gdxClose(pgdx)

        if os.path.exists(results_filename):
            ret = gdxcc.gdxOpenRead(pgdx, results_filename)
            if not ret[0]:
                raise RuntimeError("GAMS GDX failure (gdxOpenRead): %d." % ret[1])

            specVals = gdxcc.doubleArray(gdxcc.GMS_SVIDX_MAX)
            rc = gdxcc.gdxGetSpecialValues(pgdx, specVals)

            specVals[gdxcc.GMS_SVIDX_EPS] = sys.float_info.min
            specVals[gdxcc.GMS_SVIDX_UNDEF] = float("nan")
            specVals[gdxcc.GMS_SVIDX_PINF] = float("inf")
            specVals[gdxcc.GMS_SVIDX_MINF] = float("-inf")
            specVals[gdxcc.GMS_SVIDX_NA] = struct.unpack(
                ">d", bytes.fromhex("fffffffffffffffe")
            )[0]
            gdxcc.gdxSetSpecialValues(pgdx, specVals)

            i = 0
            while True:
                i += 1
                ret = gdxcc.gdxDataReadRawStart(pgdx, i)
                if not ret[0]:
                    break

                ret = gdxcc.gdxDataReadRaw(pgdx)
                if not ret[0] or len(ret[2]) < 2:
                    raise RuntimeError("GAMS GDX failure (gdxDataReadRaw).")
                level = ret[2][0]
                dual = ret[2][1]

                ret = gdxcc.gdxSymbolInfo(pgdx, i)
                if not ret[0]:
                    break
                if len(ret) < 2:
                    raise RuntimeError("GAMS GDX failure (gdxSymbolInfo).")
                model_soln[ret[1]] = (level, dual)

            gdxcc.gdxDataReadDone(pgdx)
            gdxcc.gdxClose(pgdx)

        gdxcc.gdxFree(pgdx)
        gdxcc.gdxLibraryUnload()
        return model_soln, stat_vars

    def _parse_dat_results(self, results_filename, statresults_filename):
        with open(statresults_filename, 'r') as statresults_file:
            statresults_text = statresults_file.read()

        stat_vars = dict()
        # Skip first line of explanatory text
        for line in statresults_text.splitlines()[1:]:
            items = line.split()
            try:
                stat_vars[items[0]] = float(items[1])
            except ValueError:
                # GAMS printed NA, just make it nan
                stat_vars[items[0]] = float('nan')

        with open(results_filename, 'r') as results_file:
            results_text = results_file.read()

        model_soln = dict()
        # Skip first line of explanatory text
        for line in results_text.splitlines()[1:]:
            items = line.split()
            model_soln[items[0]] = (items[1], items[2])

        return model_soln, stat_vars


class OutputStream:
    """Output stream object for simultaneously writing to multiple streams.

    tee=False:
        If set writing to this stream will write to stdout.
    logfile=None:
        Optionally a logfile can be written.

    """

    def __init__(self, tee=False, logfile=None):
        """Initialize output stream object."""
        if tee:
            self.tee = sys.stdout
        else:
            self.tee = None
        self.logfile = logfile
        self.logfile_buffer = None

    def __enter__(self):
        """Enter context of output stream and open logfile if given."""
        if self.logfile is not None:
            self.logfile_buffer = open(self.logfile, 'a')
        return self

    def __exit__(self, *args, **kwargs):
        """Enter context of output stream and close logfile if necessary."""
        if self.logfile_buffer is not None:
            self.logfile_buffer.close()
        self.logfile_buffer = None

    def write(self, message):
        """Write messages to all streams."""
        if self.tee is not None:
            self.tee.write(message)
        if self.logfile_buffer is not None:
            self.logfile_buffer.write(message)

    def flush(self):
        """Needed for python3 compatibility."""
        if self.tee is not None:
            self.tee.flush()
        if self.logfile_buffer is not None:
            self.logfile_buffer.flush()


def check_expr_evaluation(model, symbolMap, solver_io):
    try:
        # Temporarily initialize uninitialized variables in order to call
        # value() on each expression to check domain violations
        uninit_vars = list()
        for var in model.component_data_objects(Var):
            if var.value is None:
                uninit_vars.append(var)
                var.set_value(0, skip_validation=True)

        # Constraints
        for con in model.component_data_objects(Constraint, active=True):
            if con.body.is_fixed():
                continue
            check_expr(con.body, con.name, solver_io)

        # Objective
        obj = list(model.component_data_objects(Objective, active=True))
        assert len(obj) == 1, "GAMS writer can only take 1 active objective"
        obj = obj[0]
        check_expr(obj.expr, obj.name, solver_io)
    finally:
        # Return uninitialized variables to None
        for var in uninit_vars:
            var.set_value(None)


def check_expr(expr, name, solver_io):
    # Check if GAMS will encounter domain violations in presolver
    # operations at current values, which are None (0) by default
    # Used to handle log and log10 violations, for example
    try:
        value(expr)
    except (ValueError, ZeroDivisionError):
        logger.warning(
            "While evaluating model.%s's expression, GAMS solver "
            "encountered an error.\nGAMS requires that all "
            "equations and expressions evaluate at initial values.\n"
            "Ensure variable values do not violate any domains, "
            "and use the warmstart=True keyword to solve()." % name
        )
        if solver_io == 'shell':
            # For shell, there is no previous exception to worry about
            # overwriting, so raise the ValueError.
            # But for direct, the GamsExceptionExecution will be raised.
            raise


def file_removal_gams_direct(tmpdir, newdir):
    if newdir:
        shutil.rmtree(tmpdir)
    else:
        os.remove(os.path.join(tmpdir, '_gams_py_gjo0.gms'))
        os.remove(os.path.join(tmpdir, '_gams_py_gjo0.lst'))
        os.remove(os.path.join(tmpdir, '_gams_py_gdb0.gdx'))
        # .pf file is not made when DebugLevel is Off
